"""
Pairwise metrics and analytical confidence intervals.

Metrics
-------

The metrics function implemented here all have the signature::

    def metric(x : np.ndarray, y : np.ndarray) -> float

Confidence intervals
--------------------

Formulas for confidence intervals have in general been taken from from
Gilleland (2010), 10.5065/D6WD3XJM,
https://opensky.ucar.edu/islandora/object/technotes:491

Other references are cited in the docstring of the respective function.

Analytical confidence interval functions implemented here are named
``<metric>_ci``, e.g. for ``bias``, the CI function is ``bias_ci``.
The signature is be::

    def metric_ci(x : np.ndarray, y : np.ndarray, m : float,
                  alpha=0.05 : float) -> float, float

where m is the metric value that has been calculated for x and y.

Typically, you should use
:py:func:`pytesmo.metrics.confidence_intervals.with_analytical_ci` for
calculating a metric CI.
"""

# IMPORTANT: DEVELOPERS NOTES
#
# Some of the RMSD methods take ddof as keyword argument. Although there are
# some warnings in the docstrings (and also when using it) that it is
# deprecated, we want to keep it for now, in order to reproduce previous
# results.


import numpy as np
from scipy import stats
import warnings

from pytesmo.metrics._fast_pairwise import (  # noqa: F401
    bias,
    mse_bias,
    mse_var,
    mse_corr,
    mse_decomposition,
    RSS,
    _ubrmsd,
    rolling_pr_rmsd,
)


has_ci = [
    "bias",
    "ubrmsd",
    "pearson_r",
    "spearman_r",
    "kendall_tau",
]

no_ci = [
    "aad",
    "mad",
    "mse_bias",
    "msd",
    "rmsd",
    "nrmsd",
    "mse_corr",
    "mse_var",
    "nash_sutcliffe",
    "index_of_agreement",
]


def bias_ci(x, y, b, alpha=0.05):
    """
    Confidence interval for bias.

    The confidence interval is the same as the confidence interval for a mean.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.
    b : float
        bias
    alpha : float, optional
        1 - confidence level, default is 0.05

    Returns
    -------
    lower, upper : float
        Lower and upper confidence interval bounds.
    """
    n = len(x)
    delta = np.std(x - y, ddof=1) / np.sqrt(n) * stats.t.ppf(1 - alpha / 2, n - 1)
    return b - delta, b + delta


def _bias_ci_from_moments(alpha, mx, my, varx, vary, cov, n):
    # This is based on the fact that:
    # var(x - y) = var(x) + var(y) - 2*cov(x,y)
    # and therefore
    # std(x - y, ddof=1) = sqrt(var(x - y, ddof=1))
    #                    = sqrt(n/(n-1) * var(x-y))
    std = np.sqrt(n / (n - 1) * (varx + vary - 2 * cov))
    delta = std / np.sqrt(n) * stats.t.ppf(1 - alpha / 2, n - 1)
    b = mx - my
    return b - delta, b + delta


def aad(x, y):
    """
    Average (=mean) absolute deviation (AAD).

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.

    Returns
    -------
    d : float
        Mean absolute deviation.
    """
    return np.mean(np.abs(x - y))


def mad(x, y):
    """
    Median absolute deviation (MAD).

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.

    Returns
    -------
    d : float
        Median absolute deviation.
    """
    return np.median(np.abs(x - y))


def msd(x, y):
    r"""
    Mean square deviation/mean square error.

    For validation, MSD (same as MSE) is defined as

    ..math::

        MSD = \frac{1}{n}\sum\limits_{i=1}^n (x_i - y_i)^2

    MSD can be decomposed into a term describing the deviation of x and y
    attributable to non-perfect correlation (r < 1), a term depending on the
    difference in variances between x and y, and the difference in means
    between x and y (bias).

    ..math::

        MSD &= MSD_{corr} + MSD_{var} + MSD_{bias}\\
            &= 2\sigma_x\sigma_y (1-r) + (\sigma_x - \sigma_y)^2
               + (\mu_x - \mu_y)^2

    This function calculates the full MSD, the function `msd_corr`, `msd_var`,
    and `msd_bias` can be used to calculate the individual components.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector
    y : numpy.ndarray
        Second input vector

    Returns
    -------
    msd : float
        Mean square deviation
    """
    return RSS(x, y) / len(x)


def rmsd(x, y, ddof=0):
    """
    Root-mean-square deviation (RMSD).

    This is the root of MSD (see :py:func:`pytesmo.metrics.msd`).  If `x` and
    `y` have the same mean (i.e. `mean(x - y = 0`) RMSD corresponds to the
    square root of the variance of `x - y`.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.
    ddof : int, optional, DEPRECATED
        Delta degree of freedom.The divisor used in calculations is N - ddof,
        where N represents the number of elements. By default ddof is zero.
        DEPRECATED: ddof is deprecated and might be removed in future versions.

    Returns
    -------
    rmsd : float
        Root-mean-square deviation.
    """
    if ddof == 0:
        return np.sqrt(msd(x, y))
    else:
        warnings.warn(
            "ddof is deprecated and might be removed in future versions of"
            " pytesmo.",
            category=DeprecationWarning,
        )
        return np.sqrt(RSS(x, y) / (len(x) - ddof))


def nrmsd(x, y, ddof=0):
    """
    Normalized root-mean-square deviation (nRMSD).

    This is normalizes RMSD by ``max(max(x), max(y)) - min(min(x), min(y))``.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.
    ddof : int, optional
        Delta degree of freedom.The divisor used in calculations is N - ddof,
        where N represents the number of elements. By default ddof is zero.
        DEPRECATED: ddof is deprecated and might be removed in future versions.

    Returns
    -------
    nrmsd : float
        Normalized root-mean-square deviation (nRMSD).
    """
    return rmsd(x, y, ddof) / (np.max([x, y]) - np.min([x, y]))


def ubrmsd(x, y, ddof=0):
    r"""
    Unbiased root-mean-square deviation (uRMSD).

    This corresponds to RMSD with mean biases removed beforehand, that is

    ..math::

        ubRMSD = \sqrt{\frac{1}{n}\sum\limits_{i=1}^n
                           \left((x - \bar{x}) - (y - \bar{y}))^2}

    NOTE: If you are scaling the data beforehand to have zero mean bias, this
    is exactly the same as RMSD.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.
    ddof : int, optional
        Delta degree of freedom.The divisor used in calculations is N - ddof,
        where N represents the number of elements. By default ddof is zero.
        DEPRECATED: ddof is deprecated and might be removed in future versions.

    Returns
    -------
    ubrmsd : float
        Unbiased root-mean-square deviation (uRMSD).
    """
    if ddof != 0:
        warnings.warn(
            "ddof is deprecated and might be removed in future versions of"
            " pytesmo.",
            DeprecationWarning,
        )
        return np.sqrt(RSS(x - np.mean(x), y - np.mean(y)) / (len(x) - ddof))
    else:
        return _ubrmsd(x, y)


def ubrmsd_ci(x, y, ubrmsd, alpha=0.05):
    """
    Confidende interval for unbiased root-mean-square deviation (uRMSD).

    Parameters
    ----------
    x : numpy.ndarray
        First input vector
    y : numpy.ndarray
        Second input vector
    ubrmsd : float
        ubRMSD for this data
    alpha : float, optional
        1 - confidence level, default is 0.05

    Returns
    -------
    lower, upper : float
        Lower and upper confidence interval bounds.
    """
    n = len(x)
    ubMSD = ubrmsd ** 2
    lb_ubMSD = n * ubMSD / stats.chi2.ppf(1 - alpha / 2, n - 1)
    ub_ubMSD = n * ubMSD / stats.chi2.ppf(alpha / 2, n - 1)
    return np.sqrt(lb_ubMSD), np.sqrt(ub_ubMSD)


def pearson_r(x, y):
    """
    Pearson's linear correlation coefficient.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector.
    y : numpy.ndarray
        Second input vector.

    Returns
    -------
    r : float
        Pearson's correlation coefficent.

    See Also
    --------
    scipy.stats.pearsonr
    """
    return stats.pearsonr(x, y)[0]


def pearson_r_ci(x, y, r, alpha=0.05):
    """
    Confidence interval for Pearson correlation coefficient.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector
    y : numpy.ndarray
        Second input vector
    r : float
        Pearson r for this data
    alpha : float, optional
        1 - confidence level, default is 0.05

    Returns
    -------
    lower, upper : float
        Lower and upper confidence interval bounds.

    References
    ----------
    Bonett, D. G., & Wright, T. A. (2000). Sample size requirements for
    estimating Pearson, Kendall and Spearman correlations. Psychometrika,
    65(1), 23-28.
    """
    n = len(x)
    v = np.arctanh(r)
    z = stats.norm.ppf(1 - alpha / 2)
    cl = v - z / np.sqrt(n - 3)
    cu = v + z / np.sqrt(n - 3)
    return np.tanh(cl), np.tanh(cu)


def spearman_r(x, y):
    """
    Spearman's rank correlation coefficient.

    Parameters
    ----------
    x : numpy.array
        First input vector.
    y : numpy.array
        Second input vector.

    Returns
    -------
    rho : float
        Spearman correlation coefficient

    See Also
    --------
    scipy.stats.spearmenr
    """
    return stats.spearmanr(x, y)[0]


def spearman_r_ci(x, y, r, alpha=0.05):
    """
    Confidence interval for Spearman rank correlation coefficient.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector
    y : numpy.ndarray
        Second input vector
    r : float
        Spearman's r for this data
    alpha : float, optional
        1 - confidence level, default is 0.05

    Returns
    -------
    lower, upper : float
        Lower and upper confidence interval bounds.

    References
    ----------
    Bonett, D. G., & Wright, T. A. (2000). Sample size requirements for
    estimating Pearson, Kendall and Spearman correlations. Psychometrika,
    65(1), 23-28.
    """
    n = len(x)
    v = np.arctanh(r)
    z = stats.norm.ppf(1 - alpha / 2)
    # see reference for this formula
    cl = v - z * np.sqrt(1 + r ** 2 / 2) / np.sqrt(n - 3)
    cu = v + z * np.sqrt(1 + r ** 2 / 2) / np.sqrt(n - 3)
    return np.tanh(cl), np.tanh(cu)


def kendall_tau(x, y):
    """
    Wrapper for scipy.stats.kendalltau

    Parameters
    ----------
    x : numpy.array
        First input vector.
    y : numpy.array
        Second input vector.

    Returns
    -------
    tau : float
        Kendall's tau statistic

    See Also
    --------
    scipy.stats.kendalltau
    """
    return stats.kendalltau(x, y)[0]


def kendall_tau_ci(x, y, tau, alpha=0.05):
    r"""
    Confidence intervall for Kendall's rank coefficient.

    Parameters
    ----------
    x : numpy.ndarray
        First input vector
    y : numpy.ndarray
        Second input vector
    tau : float
        Kendall tau for this data
    alpha : float, optional
        1 - confidence level, default is 0.05

    Returns
    -------
    lower, upper : float
        Lower and upper confidence interval bounds.

    References
    ----------
    Bonett, D. G., & Wright, T. A. (2000). Sample size requirements for
    estimating Pearson, Kendall and Spearman correlations. Psychometrika,
    65(1), 23-28.
    """
    n = len(x)
    v = np.arctanh(tau)
    z = stats.norm.ppf(1 - alpha / 2)
    # see reference for this formula
    cl = v - z * 0.431 / np.sqrt(n - 3)
    cu = v + z * 0.431 / np.sqrt(n - 3)
    return np.tanh(cl), np.tanh(cu)


def index_of_agreement(o, p):
    """
    Index of agreement was proposed by Willmot (1981), to overcome the
    insenstivity of Nash-Sutcliffe efficiency E and R^2 to differences in the
    observed and predicted means and variances (Legates and McCabe, 1999).
    The index of agreement represents the ratio of the mean square error and
    the potential error (Willmot, 1984). The potential error in the denominator
    represents the largest value that the squared difference of each pair can
    attain. The range of d is similar to that of R^2 and lies between
    0 (no correlation) and 1 (perfect fit).

    Parameters
    ----------
    o : numpy.ndarray
        Observations.
    p : numpy.ndarray
        Predictions.

    Returns
    -------
    d : float
        Index of agreement.
    """
    denom = np.sum((np.abs(p - np.mean(o)) + np.abs(o - np.mean(o))) ** 2)
    d = 1 - RSS(o, p) / denom
    return d


def nash_sutcliffe(o, p):
    """
    Nash Sutcliffe model efficiency coefficient E. The range of E lies between
    1.0 (perfect fit) and -inf.

    Parameters
    ----------
    o : numpy.ndarray
        Observations.
    p : numpy.ndarray
        Predictions.

    Returns
    -------
    E : float
        Nash Sutcliffe model efficiency coefficient E.
    """
    return 1 - (np.sum((o - p) ** 2)) / (np.sum((o - np.mean(o)) ** 2))



import numpy as np

from pytesmo.cdf_matching import (CDFMatching,
                                  _matlab_percentile_values_from_sorted,
                                  _unique_percentile_interpolation)


def cdf_matching_testdata():
    n = 1000
    src = 1 + np.arange(n, dtype=float)
    ref = src / 2
    return src, ref


def test_cdf_matching():
    src, ref = cdf_matching_testdata()

    matcher = CDFMatching().fit(src, ref)
    assert matcher.x_perc_.shape == (101,)
    assert matcher.y_perc_.shape == (101,)

    pred = matcher.predict(src)
    np.testing.assert_almost_equal(pred, ref)

    factor = pred / src
    assert np.all(np.abs(factor - 0.5) < 1e-4)


def test_cdf_matching_percentiles():
    src, ref = cdf_matching_testdata()

    percentiles = np.linspace(0, 100, 11)
    matcher = CDFMatching(percentiles=percentiles).fit(src, ref)
    assert np.all(matcher.percentiles == percentiles)
    assert matcher.x_perc_.shape == (len(percentiles),)
    assert matcher.y_perc_.shape == (len(percentiles),)

    # make sure the result is still right
    pred = matcher.predict(src)
    np.testing.assert_almost_equal(pred, ref)
    factor = pred / src
    assert np.all(np.abs(factor - 0.5) < 1e-4)


def test_cdf_matching_resize_percentiles():
    """
    Tests whether the bin resizing works.
    """
    src, ref = cdf_matching_testdata()
    src = src[0:100]
    ref = ref[0:100]

    matcher = CDFMatching(nbins=100, minobs=20).fit(src, ref)
    # there should be only 5 bins (6 percentiles)
    assert np.sum(~np.isnan(matcher.percentiles_)) == 6
    assert np.sum(~np.isnan(matcher.x_perc_)) == 6
    assert np.sum(~np.isnan(matcher.y_perc_)) == 6


def test_cdf_matching_scaling():
    # this tests whether the scaling makes the edges of the ref CDF more robust
    # against outliers
    src, ref = cdf_matching_testdata()
    ref[0] = -9

    matcher_scaled = CDFMatching(linear_edge_scaling=True).fit(src, ref)
    matcher_unscaled = CDFMatching(linear_edge_scaling=False).fit(src, ref)
    assert matcher_scaled.x_perc_.shape == (101,)
    assert matcher_scaled.y_perc_.shape == (101,)
    assert matcher_unscaled.x_perc_.shape == (101,)
    assert matcher_unscaled.y_perc_.shape == (101,)

    pred_scaled = matcher_scaled.predict(src)
    pred_unscaled = matcher_unscaled.predict(src)

    # everything above the first bin should not be affected by the scaling
    np.all(pred_scaled[20:] == pred_unscaled[20:])

    # the unscaled transformed should have -9 as lowest value
    assert pred_unscaled[0] == -9
    # the scaled transformed value should be much closer to zero
    assert -2.5 < pred_scaled[0] < 0


def test_cdf_matching_combine_invalid():
    src, ref = cdf_matching_testdata()
    ref_orig = np.array(ref)
    ref[100:200] = np.nan

    matcher = CDFMatching(combine_invalid=True).fit(src, ref)
    assert matcher.x_perc_.shape == (101,)
    assert matcher.y_perc_.shape == (101,)

    pred = matcher.predict(src)
    np.testing.assert_almost_equal(pred, ref_orig)

    factor = pred / src
    assert np.all(np.abs(factor - 0.5) < 1e-4)


def test_cdf_matching_single_bin():
    src, ref = cdf_matching_testdata()

    matcher = CDFMatching(percentiles=[0, 100]).fit(src, ref)
    assert matcher.x_perc_.shape == (2,)
    assert matcher.y_perc_.shape == (2,)

    pred = matcher.predict(src)
    np.testing.assert_almost_equal(pred, ref)

    factor = pred / src
    assert np.all(np.abs(factor - 0.5) < 1e-4)


def test_matlab_percentile_values():
    """
    Test the percentile implementation that is used in Matlab.
    """

    arr1 = np.array([1, 1, 1, 2, 2, 2, 5, 5, 6, 10, 10, 10, 10])
    percentiles = [0, 5, 10, 20, 30, 40, 50, 60, 70, 80, 90, 95, 100]
    perc_should = [
        1.0, 1.0, 1.0, 1.1, 2.0, 2.0, 5.0, 5.3, 8.4, 10., 10., 10., 10.
    ]
    perc = _matlab_percentile_values_from_sorted(np.sort(arr1), percentiles)
    np.testing.assert_almost_equal(perc, perc_should)


def test_unique_percentile_interpolation():
    """
    test generation of unique percentile values
    by interpolation or order k
    """

    arr1 = np.array([1, 1, 1, 2, 2, 2, 5, 5, 6, 10, 10, 10, 10])
    percentiles = np.array([0, 5, 10, 20, 30, 40, 50, 60, 70, 80, 90, 95, 100])
    p = _matlab_percentile_values_from_sorted(np.sort(arr1), percentiles)
    src_perc = _unique_percentile_interpolation(p, percentiles=percentiles)
    assert len(p) == len(src_perc)

    np.testing.assert_almost_equal(src_perc, [
        1., 1.025, 1.05, 1.1, 2., 3.5, 5., 5.3, 8.4, 8.93333333, 9.46666667,
        9.73333333, 10.
    ])



# Copyright (c) 2015,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Test for climatology and anomaly calculation.
'''
import pdb

import pandas as pd
import pandas.testing as pdt
import numpy as np

import pytesmo.time_series.anomaly as anomaly


def test_anomaly_calc_given_climatology():

    clim = pd.Series(np.arange(366), name='clim', index=np.arange(366) + 1)
    data = pd.Series(
        np.arange(366), index=pd.date_range('2000-01-01', periods=366))
    anom_should = pd.Series(
        np.zeros(366), index=pd.date_range('2000-01-01', periods=366))
    anom = anomaly.calc_anomaly(
        data, climatology=clim, respect_leap_years=False)

    pdt.assert_series_equal(anom_should, anom, check_dtype=False)


def test_anomaly_calc_given_climatology_return_clim():

    clim = pd.Series(np.arange(366), name='clim', index=np.arange(366) + 1)
    data = pd.Series(
        np.arange(366), index=pd.date_range('2000-01-01', periods=366))
    anom_should = pd.Series(
        np.zeros(366), index=pd.date_range('2000-01-01', periods=366),
        name='anomaly')
    anom = anomaly.calc_anomaly(
        data, climatology=clim, respect_leap_years=False,
        return_clim=True)

    pdt.assert_series_equal(anom_should, anom['anomaly'], check_dtype=False)


def test_anomaly_calc_given_climatology_no_leap_year():

    clim = pd.Series(np.arange(366), name='clim', index=np.arange(366) + 1)
    data = pd.Series(
        np.arange(365), index=pd.date_range('2007-01-01', periods=365))
    anom_should = pd.Series(
        np.zeros(365), index=pd.date_range('2007-01-01', periods=365))
    anom = anomaly.calc_anomaly(
        data, climatology=clim, respect_leap_years=True)

    pdt.assert_series_equal(anom_should, anom, check_dtype=False)


def test_climatology_always_366():
    ts = pd.Series(np.sin(np.arange(366) / 366. * 2 * np.pi), index=pd.date_range(
        '2000-01-01', freq='D', periods=366))
    # remove a part of the time series
    ts['2000-02-01': '2000-02-28'] = np.nan
    ts = ts.dropna()
    clim = anomaly.calc_climatology(ts)
    assert clim.size == 366

    # this should also be the case if interpolate_leapday is set
    ts = pd.Series(np.sin(np.arange(10)), index=pd.date_range(
        '2000-01-01', freq='D', periods=10))
    clim = anomaly.calc_climatology(ts, interpolate_leapday=True)
    assert clim.size == 366


def test_monthly_climatology_always_12():
    ts = pd.Series(np.sin(np.arange(366) / 366. * 2 * np.pi), index=pd.date_range(
        '2000-01-01', freq='D', periods=366))
    # remove a part of the time series
    ts['2000-02-01': '2000-02-28'] = np.nan
    ts = ts.dropna()
    clim = anomaly.calc_climatology(ts, output_freq="month")
    assert clim.size == 12

    # this should also be the case if interpolate_leapday is set
    ts = pd.Series(np.sin(np.arange(10)), index=pd.date_range(
        '2000-01-01', freq='D', periods=10))
    # and if the window size is changed
    clim = anomaly.calc_climatology(ts, moving_avg_clim=5, output_freq="month")
    assert clim.size == 12


def test_climatology_always_366_fill():
    ts = pd.Series(np.sin(np.arange(366) / 366. * 2 * np.pi), index=pd.date_range(
        '2000-01-01', freq='D', periods=366))
    # remove a part of the time series
    ts['2000-02-01': '2000-02-28'] = np.nan
    ts = ts.dropna()
    clim = anomaly.calc_climatology(ts, fill=-1, moving_avg_clim=1)
    assert clim.size == 366
    assert clim.iloc[31] == -1


def test_climatology_closed():
    ts = pd.Series(np.arange(366), index=pd.date_range(
        '2000-01-01', freq='D', periods=366))
    # remove a part of the time series
    ts['2000-02-01': '2000-02-28'] = np.nan
    ts = ts.dropna()
    clim = anomaly.calc_climatology(ts, wraparound=True)
    assert clim.size == 366
    # test that the arange was closed during the second moving average
    assert clim.iloc[365] - 187.90 < 0.01



import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when
    # `python_requires = >= 3.8`
    from importlib.metadata import (
        PackageNotFoundError,
        version,
    )  # pragma: no cover
else:
    from importlib_metadata import (
        PackageNotFoundError,
        version,
    )  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError


# custom code not generated by PyScaffold
__all__ = [
    "metrics",
    "scaling",
    "temporal_matching",
    "timedate",
    "time_series",
    "grid",
    "colormaps",
    "validation_framework"
]



"""
Provides functions for temporally collocating data from multiple dataframes.
"""

from numba import guvectorize, float32, float64
import numpy as np
import pandas as pd
from pykdtree.kdtree import KDTree
import warnings

__all__ = ["temporal_collocation", "combined_temporal_collocation"]


def df_match(reference, *args, **kwds):
    """
    **Deprecated!**

    Finds temporal match between the reference pandas.DataFrame (index has to
    be datetime) and n other pandas.DataFrame (index has to be datetime).

    Parameters
    ----------
    reference : pandas.DataFrame or pandas.TimeSeries
        The index of this dataframe will be the reference.
    *args : pandas.DataFrame or pandas.TimeSeries
        The index of this dataframe(s) will be matched.
        If it is a pandas.Series then it has to have a name. Otherwise
        no column name can be assigned to the matched DataFrame.
    window : float
        Fraction of days of the maximum pos./neg. distance allowed, i.e. the
        value of window represents the half-winow size (e.g. window=0.5, will
        search for matches between -12 and +12 hours) (default: None)
    dropna : boolean
        Drop rows containing only NaNs (default: False)
    dropduplicates : boolean
        Drop duplicated temporal matched (default: False)
    asym_window: string, optional
        ``<=`` stands for using a smaller and equal only for the left/smaller
        side of the window comparison, ``>=`` stands for using a larger and
        equal only for the right/larger side of the window comparison.
        The default is to use <= and >= for both sides of the search window

    Returns
    -------
    temporal_matched_args : pandas.DataFrame or tuple of pandas.DataFrame
        Dataframe with index from matched reference index
    """
    warnings.warn(
        "'pytesmo.temporal_matching.df_match' is deprecated. Use"
        "'pytesmo.temporal_matching.temporal_collocation' instead!",
        DeprecationWarning,
    )

    if "window" in kwds:
        window = kwds["window"]
    else:
        window = None

    if "asym_window" in kwds:
        asym_window = kwds["asym_window"]
    else:
        asym_window = None

    temporal_matched_args = []
    ref_step = reference.index.values - reference.index.values[0]

    for arg in args:

        if type(arg) is pd.Series:
            arg = pd.DataFrame(arg)
        comp_step = arg.index.values - reference.index.values[0]
        values = np.arange(comp_step.size)
        # setup kdtree which must get 2D input
        try:
            tree = KDTree(np.atleast_2d(comp_step).T, balanced_tree=False)
        except TypeError:
            # scipy before version 0.16 does not have the balanced_tree kw
            # but is fast in this case also without it
            tree = KDTree(np.atleast_2d(comp_step).T)

        dist, i = tree.query(np.atleast_2d(ref_step).T)
        matched = values[i]

        distance = np.zeros_like(matched, dtype=float)
        distance.fill(np.nan)
        valid_match = np.invert(np.isnan(matched))

        distance[valid_match] = (
            arg.index.values[np.int32(matched[valid_match])] -
            reference.index.values[valid_match]) / np.timedelta64(1, "D")

        arg = arg.assign(index=arg.index.values, merge_key=np.arange(len(arg)))

        arg_matched = pd.DataFrame({
            "merge_key": matched,
            "distance": distance,
            "ref_index": reference.index.values,
        })
        arg_matched = arg_matched.merge(arg, on="merge_key", how="left")
        arg_matched.index = arg_matched["ref_index"].values
        arg_matched = arg_matched.sort_index()

        if window is not None:
            if asym_window is None:
                invalid_dist = arg_matched["distance"].abs() > window
            if asym_window == "<=":
                # this means that only distance in the interval [distance[ are
                # taken
                valid_dist = ((arg_matched["distance"] >= 0.0)
                              & (arg_matched["distance"] <= window)) | (
                                  (arg_matched["distance"] <= 0.0)
                                  & (arg_matched["distance"] > -window))
                invalid_dist = ~valid_dist
            if asym_window == ">=":
                # this means that only distance in the interval ]distance] are
                # taken
                valid_dist = ((arg_matched["distance"] >= 0.0)
                              & (arg_matched["distance"] < window)) | (
                                  (arg_matched["distance"] <= 0.0)
                                  & (arg_matched["distance"] >= -window))
                invalid_dist = ~valid_dist
            arg_matched.loc[invalid_dist] = np.nan

        if "dropna" in kwds and kwds["dropna"]:
            arg_matched = arg_matched.dropna(how="all")

        if "dropduplicates" in kwds and kwds["dropduplicates"]:
            arg_matched = arg_matched.dropna(how="all")
            g = arg_matched.groupby("merge_key")
            min_dists = g.distance.apply(lambda x: x.abs().idxmin())
            arg_matched = arg_matched.loc[min_dists]

        temporal_matched_args.append(
            arg_matched.drop(["merge_key", "ref_index"], axis=1))

    if len(temporal_matched_args) == 1:
        return temporal_matched_args[0]
    else:
        return tuple(temporal_matched_args)


def matching(reference, *args, **kwargs):
    """
    Finds temporal match between the reference pandas.TimeSeries (index has to
    be datetime) and n other pandas.TimeSeries (index has to be datetime).

    Parameters
    ----------
    reference : pandas.TimeSeries
        The index of this Series will be the reference.
    *args : pandas.TimeSeries
        The index of these Series(s) will be matched.
    window : float
        Fraction of days of the maximum pos./neg. distance allowed, i.e. the
        value of window represents the half-winow size (e.g. window=0.5, will
        search for matches between -12 and +12 hours) (default: None)

    Returns
    -------
    temporal_match : pandas.DataFrame
        containing the index of the reference Series and a column for each of
        the other input Series
    """
    warnings.warn(
        "'pytesmo.temporal_matching.matching' is deprecated. Use"
        "'pytesmo.temporal_matching.temporal_collocation' instead!",
        DeprecationWarning,
    )
    matched_datasets = df_match(
        reference, *args, dropna=True, dropduplicates=True, **kwargs)

    if type(matched_datasets) != tuple:
        matched_datasets = [matched_datasets]

    matched_data = pd.DataFrame(reference)

    for match in matched_datasets:
        match = match.drop(["distance", "index"], axis=1)
        matched_data = matched_data.join(match)

    return matched_data.dropna()


def temporal_collocation(
    reference,
    other,
    window,
    method="nearest",
    return_index=False,
    return_distance=False,
    dropduplicates=False,
    dropna=False,
    checkna=False,
    flag=None,
    use_invalid=False,
):
    """
    Temporally collocates values to reference.

    Parameters
    ----------
    reference : pd.DataFrame, pd.Series, or pd.DatetimeIndex
        The reference onto which `other` should be collocated. If this is a
        DataFrame or a Series, the index must be a DatetimeIndex. If the index
        is timezone-naive and `other` is not, the timezone of `other` will be
        assumed.
    other : pd.DataFrame or pd.Series
        Data to be collocated. Must have a pd.DatetimeIndex as index. If the
        index is timezone-naive and `reference` is not, the timezone of the
        reference data will be assumed.
    window : pd.Timedelta or float
        Window around reference timestamps in which to look for data. Floats
        are interpreted as number of days.
    method : str, optional
        Which method to use for the temporal collocation:

        - "nearest" (default): Uses the nearest valid neighbour. When this
          method is used, entries with duplicate index values in `other` will
          be dropped, and only the first of the duplicates is kept.
        - "mean": Takes the mean over the given window around the reference
          times.

    return_index : boolean, optional
        Include index of `other` in matched dataframe (default: False). Only
        used with ``method="nearest"``. The index will be added as a separate
        column with the name "index_other".
    return_distance : boolean, optional
        Include distance information between `reference` and `other` in matched
        dataframe (default: False). This is only used with
        ``method="nearest"``, and implies ``return_index=True``. The distance
        will be added as a separate column with the name "distance_other".
    dropduplicates : bool, optional
        Whether to drop duplicated timestamps in `other`. Default is ``False``,
        except when ``method="nearest"``, in which case this is enforced to be
        ``True``.
    dropna : bool, optional
        Whether to drop NaNs from the resulting dataframe (arising for example
        from duplicates with ``duplicates_nan=True`` or from missing values).
        This uses ``how="all"``, that is, only rows where all values are NaN
        are dropped. Default is ``False``.
    checkna: bool, optional
        Whether to check if only NaNs are returned (i.e. no match has been
        found). If set to ``True``, raises a ``UserWarning`` in case no match
        has been found. Default is ``False``.
    flag : np.ndarray, str or None, optional
        Flag column as array or name of the flag column in `other`. If this is
        given, the column will be interpreted as validity indicator. Any
        nonzero values mark the row as invalid. Default is ``None``.
    use_invalid : bool, optional
        Whether to use invalid values marked by `flag` in case no valid values
        are available. Default is ``False``.

    Returns
    -------
    collocated : pd.DataFrame or pd.Series
        Temporally collocated version of ``other``.
    """

    # input validation
    # ----------------
    if isinstance(reference, (pd.Series, pd.DataFrame)):
        ref_dr = reference.index
    elif isinstance(reference, pd.DatetimeIndex):
        ref_dr = reference
    else:  # pragma: no cover
        raise ValueError(
            "'reference' must be pd.DataFrame, pd.Series, or pd.DatetimeIndex."
        )
    if not isinstance(other, (pd.Series, pd.DataFrame)):  # pragma: no cover
        raise ValueError("'other' must be pd.DataFrame or pd.Series.")
    if not isinstance(window, pd.Timedelta):
        window = pd.Timedelta(days=window)
    if flag is not None:
        if isinstance(flag, str):
            flag = other[flag].values
        if len(flag) != len(ref_dr):  # pragma: no cover
            raise ValueError("Flag must have same length as reference")
        flagged = flag.astype(bool)
        has_invalid = np.any(flagged)
    else:
        has_invalid = False

    # preprocessing
    # ------------
    if ref_dr.tz is None and other.index.tz is None:
        # no timezone info provided for any of the inputs, so we will continue
        # to use timezone naive frames
        pass
    else:
        if ref_dr.tz is None:
            ref_dr = ref_dr.tz_localize(other.index.tz)
            warnings.warn(
                "No timezone given for reference, assuming it's in the same"
                f" timezone as other, {other.index.tz}.",
                UserWarning,
            )
        elif other.index.tz is None:
            other = other.tz_localize(ref_dr.tz)
            warnings.warn(
                "No timezone given for other, assuming it's in the same"
                f" timezone as reference, {other.index.tz}.",
                UserWarning,
            )
        if other.index.tz != ref_dr.tz:
            other = other.tz_convert(ref_dr.tz)

    if dropduplicates or method == "nearest":
        other = other[~other.index.duplicated(keep="first")]
        ref_duplicated = ref_dr.duplicated(keep="first")
        if np.any(ref_duplicated):
            warnings.warn("Dropping duplicated indices in reference."
                          " This might indicate issues with your data.")
            ref_dr = ref_dr[~ref_dr.duplicated(keep="first")]

    # collocation
    # -----------
    if method == "nearest":
        # Nearest neighbour collocation, uses pandas reindex

        if return_index or return_distance:
            new_cols = {}
            new_cols["index_other"] = other.index
            if return_distance:
                new_cols["distance_other"] = np.zeros(len(other))
            other = other.assign(**new_cols)

        def collocate(df):
            return df.reindex(ref_dr, method="nearest", tolerance=window)

        if has_invalid:
            collocated = collocate(other[~flagged])
            if use_invalid:
                invalid = collocate(other[flagged])
                collocated = collocated.combine_first(invalid)
        else:
            collocated = collocate(other)

        if return_distance:
            collocated["distance_other"] = (
                collocated["index_other"] - collocated.index)

    elif method == "mean":

        window_days = 2 * window / pd.Timedelta(1, "D")
        other_times = other.index.to_julian_date().values
        if not has_invalid or use_invalid:
            mask = np.ones_like(other_times, dtype=bool)
        else:
            mask = ~flagged

        other_is_series = isinstance(other, pd.Series)
        if other_is_series:
            other = pd.DataFrame(other, columns=[other.name])

        ncols = other.shape[1]
        data = np.empty((ncols, len(ref_dr)), dtype=other.iloc[:, 0].dtype)
        ref_dr_jd = ref_dr.to_julian_date().values
        for i in range(ncols):
            other_data = other.iloc[:, i].values[mask]
            data[i, :] = resample_mean(
                other_times, other_data, ref_dr_jd, window_days
            )
        collocated = pd.DataFrame(data.T, index=ref_dr, columns=other.columns)

        if other_is_series:
            collocated = collocated.iloc[:, 0]

    else:
        raise NotImplementedError(
            "Only nearest neighbour collocation is implemented so far")

    # postprocessing
    # --------------
    if checkna:
        if np.any(collocated.isnull().apply(np.all)):
            warnings.warn("No match has been found")
    if dropna:
        collocated.dropna(inplace=True, how="all")

    return collocated


def combined_temporal_collocation(
    reference,
    others,
    window,
    method="nearest",
    dropduplicates=False,
    dropna=False,
    combined_dropna=False,
    flag=None,
    checkna=False,
    use_invalid=False,
    add_ref_data=False,
):
    """
    Temporally collocates multiple dataframes to reference times.

    Parameters
    ----------
    reference : pd.DataFrame, pd.Series, or pd.DatetimeIndex
        The reference onto which `other` should be collocated. If this is a
        DataFrame or a Series, the index must be a DatetimeIndex. If the index
        is timezone-naive, UTC will be assumed.
    others : list/tuple of pd.DataFrame or pd.Series
        DataFrames/Series to be collocated. Each entry must have a
        pd.DatetimeIndex as index. If the index is timezone-naive, the timezone
        of the reference data will be assumed.
    window : pd.Timedelta or float
        Window around reference timestamps in which to look for data. Floats
        are interpreted as number of days.
    method : str, optional
        Which method to use for the temporal collocation:

        - "nearest" (default): Uses the nearest valid neighbour. When this
          method is used, entries with duplicate index values in `other` will
          be dropped, and only the first of the duplicates is kept.
        - "mean": Takes the mean over the given window around the reference
          times.

    dropduplicates : bool, optional
        Whether to drop duplicated timestamps in `others`. Default is
        ``False``, except when ``method="nearest"``, in which case this is
        enforced to be ``True``.
    dropna : bool, optional
        Whether to drop NaNs from the resulting dataframe (arising for example
        from duplicates with ``duplicates_nan=True`` or from missing values).
        Default is ``False``.
    combined_dropna : str or bool, optional
        Whether and how to drop NaNs from the resulting combined DataFrame. Can
        be ``"any"``, ``"all"``, ``True`` or ``False``. "any" makes sure that
        the output dataframe only has values at times where all input frames
        had values, while "all" only drops lines where all values are NaN.
        ``True`` is the same as "any", and ``False`` (default) disables
        dropping NaNs.
    checkna: bool, optional
        Whether to check if only NaNs are returned (i.e. no match has been
        found). If set to ``True``, raises a ``UserWarning`` in case no match
        has been found. Default is ``False``.
    flag : np.ndarray or None, optional
        Flag column as array. If this is given, the column will be interpreted
        as validity indicator. Any nonzero values mark the row as invalid.
        Default is ``None``.
    use_invalid : bool, optional
        Whether to use invalid values marked by `flag` in case no valid values
        are available. Default is ``False``.
    add_ref_data : bool, optional
        If `reference` is a DataFrame or Series, add the data to the final
        collocated dataframe.

    Returns
    -------
    collocated : pd.DataFrame or pd.Series
        Temporally collocated DataFrame with variables from all input frames
        merged together.
    """
    dfs = [
        temporal_collocation(
            reference,
            other,
            window,
            method=method,
            return_index=False,
            return_distance=False,
            dropduplicates=dropduplicates,
            dropna=dropna,
            checkna=checkna,
            flag=flag,
            use_invalid=use_invalid,
        ) for other in others
    ]
    if isinstance(reference, (pd.DataFrame, pd.Series)) and add_ref_data:
        # first, check if we have to remove duplicates
        if dropduplicates:
            duplicated = reference.index.duplicated(keep="first")
            reference = reference[~duplicated]
        dfs.insert(0, reference)

    # Before merging we have to check if the timezones are consistent.
    timezones = [d.index.tz for d in dfs]
    uniq_tzs = set(timezones)
    if len(uniq_tzs) == 1:
        pass
    else:
        actual_tzs = set(tz for tz in timezones if tz is not None)
        if len(actual_tzs) == 1:
            # convert all to common tz
            tz = list(actual_tzs)[0]
        else:
            # Multiple different timezones, convert all to UTC and raise a
            # warning
            tz = "UTC"
            warnings.warn(
                "Input DataFrames have mixed timezones, converting everything"
                " to UTC.",
                UserWarning,
            )
        for d in dfs:
            if d.index.tz is None:
                d.index = d.index.tz_localize(tz)
            else:
                d.index = d.index.tz_convert(tz)

    merged = pd.concat(dfs, axis=1)
    if combined_dropna:
        if combined_dropna is True:
            combined_dropna = "any"
        merged = merged.dropna(how=combined_dropna)
    return merged


@guvectorize(
    [
        (float32[:], float32[:], float32[:], float32, float32[:]),
        (float64[:], float64[:], float64[:], float64, float64[:]),
    ],
    "(n), (n), (m), () -> (m)",
    nopython=True,
)
def resample_mean(times, values, target_times, window,
                  resampled):  # pragma: no cover
    """
    Resamples to new times by taking a mean over a given window.

    Parameters
    ----------
    times : np.ndarray
        Times at which values are taken as float.
    values : np.ndarray
        Array with values.
    target_times : np.ndarray
        New times to which to resample.
    window : float
        Size of the window as float, in the same units as the times (e.g., if
        the times are in units of days, this should be in days).

    Returns
    -------
    resampled : np.ndarray
    """
    n_orig = len(times)
    n_target = len(target_times)
    lower = 0
    upper = 0
    half_window = window / 2.0

    for i in range(n_target):
        for j in range(lower, n_orig + 1):
            lower = j
            if (
                lower == n_orig
                or times[j] >= target_times[i] - half_window
            ):
                break
        # check if the current window is still below the last time that we have
        if times[n_orig - 1] > target_times[i] + half_window:
            for j in range(max(0, upper), n_orig):
                upper = j - 1
                if times[j] > target_times[i] + half_window:
                    break
        else:
            upper = n_orig - 1

        nobs = max(upper - lower + 1, 0)
        if nobs == 0:
            resampled[i] = np.nan
        else:
            resampled[i] = np.nanmean(values[lower:(upper + 1)])



# cython: boundscheck=False, wraparound=False, cdivision=True, nonecheck=False
import numpy as np
cimport numpy as cnp
from numpy.math cimport NAN
cimport cython
from cython cimport floating
from libc.math cimport sqrt, fabs
from scipy.special.cython_special import betainc


cpdef _moments_welford_nd(floating [:,:] X):
    """
    Calculates means, variances, and covariance of the given input array using
    Welford's algorithm.

    Parameters
    ----------
    X : numpy.ndarray
        num_samples x num_datasets data array

    Returns
    -------
    mean : numpy.ndarray
        Array of means (length num_datasets)
    cov : numpy.ndarray
        Covariance matrix (num_datasets x num_datasets)
    """
    cdef int i, j, k, n, d
    cdef floating nobs
    cdef floating [:] mean
    cdef floating [:] oldmean
    cdef floating [:,:] C
    n = X.shape[0]
    d = X.shape[1]

    if floating is float:
        dtype = np.float32
    else:
        dtype = np.float64

    mean = np.zeros(d, dtype=dtype)
    oldmean = np.zeros(d, dtype=dtype)
    C = np.zeros((d, d), dtype=dtype)

    nobs = 0
    for i in range(n):
        nobs += 1
        for j in range(d):
            oldmean[j] = mean[j]
            mean[j] += (X[i, j] - mean[j]) / nobs
            for k in range(j, d):
                C[j, k] += (X[i, j] - mean[j]) * (X[i, k] - mean[k])

    for j in range(d):
        for k in range(j, d):
            # divide C by nobs to get covariance
            C[j, k] /= nobs
            # populate lower half of C
            if k != j:
                C[k, j] = C[j, k]

    return mean, C



# -*- coding: utf-8 -*-

import warnings
warnings.warn("pytesmo.timdate.dekad has moved to the cadati pip package "
              "Import from pytesmo will not work in future versions of the package",
              DeprecationWarning)

from cadati.dekad import *


# -*- coding: utf-8 -*-
# <nbformat>3.0</nbformat>

# <codecell>

import ascat.timeseries as ascat
import pytesmo.time_series.anomaly as ts_anomaly

import os
import matplotlib.pyplot as plt

testdata_folder = '/pytesmo/testdata'

# <codecell>

ascat_data_folder = os.path.join(testdata_folder,
                                 'sat/ascat/netcdf/55R22')
ascat_grid_folder = os.path.join(testdata_folder,
                                 'sat/ascat/netcdf/grid')
static_layers_folder = os.path.join(testdata_folder,
                                    'sat/h_saf/static_layer')

#init the ASCAT_SSM reader with the paths
ascat_SSM_reader = ascat.AscatSsmCdr(ascat_data_folder, ascat_grid_folder,
                                     grid_filename='TUW_WARP5_grid_info_2_1.nc',
                                     static_layer_path=static_layers_folder)

# <codecell>

# Read data for location in northern Italy
ascat_ts = ascat_SSM_reader.read(11, 45)
#plot soil moisture
ascat_ts.data['sm'].plot(title='SSM data')
plt.show()

# <codecell>

#calculate anomaly based on moving +- 17 day window
anom = ts_anomaly.calc_anomaly(ascat_ts.data['sm'], window_size=35)
anom.plot(title='Anomaly (35-day window)')
plt.show()

# <codecell>

#calculate climatology
climatology = ts_anomaly.calc_climatology(ascat_ts.data['sm'])
climatology.plot(title='Climatology')
plt.show()

# <codecell>

#calculate anomaly based on climatology
anomaly_clim = ts_anomaly.calc_anomaly(ascat_ts.data['sm'], climatology=climatology)
anomaly_clim.plot(title='Anomaly (climatology)')
plt.show()

# <codecell>



# Copyright (c) 2020, TU Wien, Department of Geodesy and Geoinformation.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of TU Wien, Department of Geodesy and Geoinformation nor
#     the names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT
# OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Metric calculators implement combinations of metrics and structure the output.
"""

import copy
import itertools
import numpy as np
import pandas as pd
from scipy import stats
import warnings

import pytesmo.metrics as metrics
import pytesmo.df_metrics as df_metrics
from pytesmo.scaling import scale
from pytesmo.validation_framework.data_manager import get_result_names
import pytesmo.validation_framework.error_handling as eh
from pytesmo.df_metrics import n_combinations
from pytesmo.metrics import (
    pairwise,
    tcol_metrics,
)
from pytesmo.metrics.confidence_intervals import (
    tcol_metrics_with_bootstrapped_ci,
    with_bootstrapped_ci,
)
from pytesmo.metrics._fast_pairwise import (
    _moments_welford,
    _mse_corr_from_moments,
    _mse_var_from_moments,
    _mse_bias_from_moments,
    _pearsonr_from_moments,
)
from pytesmo.metrics.pairwise import (
    _bias_ci_from_moments,
    ubrmsd_ci,
    pearson_r_ci,
    spearman_r_ci,
    kendall_tau_ci,
)


def _get_tc_metric_template(metr, ds_names):
    """return empty dict to fill TC results into"""
    if isinstance(metr, str):
        metr = [metr]

    # metrics that are computed between dataset triples
    met_thds_template = {
        "snr": np.float32([np.nan]),
        "err_std": np.float32([np.nan]),
        "beta": np.float32([np.nan]),
    }

    in_lut = np.isin(
        np.array(metr), np.array(list(met_thds_template.keys()))
    )
    if not all(in_lut):
        unknown = np.take(metr, np.where(in_lut is False)[0])
        raise ValueError("Unknown metric(s): {}".format(" ".join(unknown)))

    met_thds = {}
    for m in metr:
        for d in ds_names:
            met_thds[(m, d)] = copy.copy(met_thds_template[m])

    return met_thds


def _get_metric_template(metr):
    """return empty dict to fill results into"""
    if isinstance(metr, str):
        metr = [metr]

    lut = dict()
    # metrics that are equal for all data sets
    met_comm = {"n_obs": np.int32([0])}
    # metrics that are computed between dataset pairs
    met_tds = {
        "R": np.float32([np.nan]),
        "p_R": np.float32([np.nan]),
        "rho": np.float32([np.nan]),
        "p_rho": np.float32([np.nan]),
        "BIAS": np.float32([np.nan]),
        "tau": np.float32([np.nan]),
        "p_tau": np.float32([np.nan]),
        "RMSD": np.float32([np.nan]),
        "mse": np.float32([np.nan]),
        "RSS": np.float32([np.nan]),
        "mse_corr": np.float32([np.nan]),
        "mse_bias": np.float32([np.nan]),
        "mse_var": np.float32([np.nan]),
        "urmsd": np.float32([np.nan]),
        "BIAS_ci_lower": np.float32([np.nan]),
        "BIAS_ci_upper": np.float32([np.nan]),
        "RMSD_ci_lower": np.float32([np.nan]),
        "RMSD_ci_upper": np.float32([np.nan]),
        "RSS_ci_lower": np.float32([np.nan]),
        "RSS_ci_upper": np.float32([np.nan]),
        "mse_ci_lower": np.float32([np.nan]),
        "mse_ci_upper": np.float32([np.nan]),
        "mse_corr_ci_lower": np.float32([np.nan]),
        "mse_corr_ci_upper": np.float32([np.nan]),
        "mse_bias_ci_lower": np.float32([np.nan]),
        "mse_bias_ci_upper": np.float32([np.nan]),
        "mse_var_ci_lower": np.float32([np.nan]),
        "mse_var_ci_upper": np.float32([np.nan]),
        "urmsd_ci_lower": np.float32([np.nan]),
        "urmsd_ci_upper": np.float32([np.nan]),
        "R_ci_lower": np.float32([np.nan]),
        "R_ci_upper": np.float32([np.nan]),
        "rho_ci_lower": np.float32([np.nan]),
        "rho_ci_upper": np.float32([np.nan]),
        "tau_ci_lower": np.float32([np.nan]),
        "tau_ci_upper": np.float32([np.nan]),
    }

    lut.update(met_comm)
    lut.update(met_tds)

    in_lut = np.isin(np.array(metr), np.array(list(lut.keys())))
    if not all(in_lut):
        unknown = np.take(metr, np.where(in_lut is False)[0])
        raise ValueError("Unknown metric(s): {}".format(" ".join(unknown)))

    return {m: lut[m] for m in metr}


class MetadataMetrics(object):
    """
    This class sets up the gpi info and metadata (if used) in the results
    template.
    This is used as the basis for all other metric calculators.

    Parameters
    ----------
    other_name: string or tuple, optional
        Name of the column of the non-reference / other dataset in the
        pandas DataFrame
    metadata_template: dictionary, optional
        A dictionary containing additional fields (and types) of the form
        dict = {'field': np.float32([np.nan]}. Allows users to specify
        information in the job tuple, i.e. jobs.append((idx,
        metadata['longitude'], metadata['latitude'], metadata_dict)) which
        is then propagated to the end netCDF results file.
    min_obs : int, optional
        Minium number of observations required t calculate metrics. Default is
        10.
    """

    def __init__(self, other_name="k1", metadata_template=None, min_obs=10):

        self.result_template = {
            "gpi": np.int32([-1]),
            "lon": np.float64([np.nan]),
            "lat": np.float64([np.nan]),
            "status": np.int32([eh.UNCAUGHT]),
        }

        self.metadata_template = metadata_template
        if self.metadata_template is not None:
            self.result_template.update(metadata_template)

        self.other_name = other_name
        self.min_obs = min_obs

    def calc_metrics(self, data, gpi_info):
        """
        Adds the gpi info and metadata to the results.

        Parameters
        ----------
        data : pandas.DataFrame
            see individual calculators for more information. not directly used
            here.
        gpi_info : tuple
            of (gpi, lon, lat) or, optionally, (gpi, lon, lat, metadata) where
            metadata is a dictionary
        """

        dataset = copy.deepcopy(self.result_template)

        dataset["gpi"][0] = gpi_info[0]
        dataset["lon"][0] = gpi_info[1]
        dataset["lat"][0] = gpi_info[2]

        if self.metadata_template is not None:
            for key, value in self.metadata_template.items():
                try:
                    dataset[key][0] = gpi_info[3][key]
                except (IndexError):
                    raise Exception(
                        "No metadata has been provided to the job. "
                        "Should be of form {field: metadata_value} using "
                        "the metadata_template supplied to init function."
                    )

        return dataset


class BasicMetrics(MetadataMetrics):
    """
    This class just computes the basic metrics,
        - Pearson's R
        - Spearman's rho
        - RMSD
        - BIAS
        - optionally Kendall's tau

    it also stores information about gpi, lat, lon
    and number of observations

    Parameters
    ----------
    other_name: string or tuple, optional
        Name of the column of the non-reference / other dataset in the
        pandas DataFrame
    calc_tau: boolean, optional
        if True then also tau is calculated. This is set to False by default
        since the calculation of Kendalls tau is rather slow and can
        significantly impact performance of e.g. global validation studies
    metadata_template: dictionary, optional
        A dictionary containing additional fields (and types) of the form
        ``dict = {'field': np.float32([np.nan]}``. Allows users to specify
        information in the job tuple, i.e.::

            jobs.append((idx, metadata['longitude'], metadata['latitude'],
                         metadata_dict))``

        which is then propagated to the end netCDF results file.
    """

    def __init__(
        self, other_name="k1", calc_tau=False, metadata_template=None
    ):
        super(BasicMetrics, self).__init__(
            other_name=other_name, metadata_template=metadata_template
        )

        self.basic_metrics = [
            "n_obs",
            "R",
            "p_R",
            "rho",
            "p_rho",
            "RMSD",
            "BIAS",
            "tau",
            "p_tau",
        ]

        self.result_template.update(_get_metric_template(self.basic_metrics))

        self.calc_tau = calc_tau

    def calc_metrics(self, data, gpi_info):
        """
        calculates the desired statistics

        Parameters
        ----------
        data : pandas.DataFrame
            with 2 columns, the first column is the reference dataset
            named 'ref'
            the second column the dataset to compare against named 'other'
        gpi_info : tuple
            of (gpi, lon, lat)

        Notes
        -----
        Kendall tau is calculation is optional at the moment
        because the scipy implementation is very slow which is problematic for
        global comparisons

        """
        dataset = super(BasicMetrics, self).calc_metrics(data, gpi_info)

        if len(data) < self.min_obs:
            dataset["status"][0] = eh.INSUFFICIENT_DATA
            return dataset

        x, y = data["ref"].values, data[self.other_name].values
        R, p_R = stats.pearsonr(x, y)
        rho, p_rho = stats.spearmanr(x, y)
        RMSD = metrics.rmsd(x, y)
        BIAS = metrics.bias(x, y)

        dataset["R"][0], dataset["p_R"][0] = R, p_R
        dataset["rho"][0], dataset["p_rho"][0] = rho, p_rho
        dataset["RMSD"][0] = RMSD
        dataset["BIAS"][0] = BIAS
        dataset["n_obs"][0] = len(data)

        if self.calc_tau:
            tau, p_tau = metrics.kendalltau(x, y)
            dataset["tau"][0], dataset["p_tau"][0] = tau, p_tau

        dataset["status"][0] = eh.OK
        return dataset


class BasicMetricsPlusMSE(BasicMetrics):
    """
    Basic Metrics plus Mean squared Error and the decomposition of the MSE
    into correlation, bias and variance parts.
    """

    def __init__(self, other_name="k1", metadata_template=None):
        super(BasicMetricsPlusMSE, self).__init__(
            other_name=other_name, metadata_template=metadata_template
        )

        self.mse_metrics = ["mse", "mse_corr", "mse_bias", "mse_var"]
        self.result_template.update(_get_metric_template(self.mse_metrics))

    def calc_metrics(self, data, gpi_info):
        dataset = super(BasicMetricsPlusMSE, self).calc_metrics(
            data, gpi_info
        )
        # setting back to uncaught in case something goes wrong
        dataset["status"][0] = eh.UNCAUGHT
        if len(data) < self.min_obs:
            dataset["status"][0] = eh.INSUFFICIENT_DATA
            return dataset
        x, y = data["ref"].values, data[self.other_name].values
        mse, mse_corr, mse_bias, mse_var = metrics.mse_decomposition(x, y)
        dataset["mse"][0] = mse
        dataset["mse_corr"][0] = mse_corr
        dataset["mse_bias"][0] = mse_bias
        dataset["mse_var"][0] = mse_var
        dataset["status"][0] = eh.OK
        return dataset


class FTMetrics(MetadataMetrics):
    """
    This class computes Freeze/Thaw Metrics
    Calculated metrics are:

    - SSF frozen/temp unfrozen
    - SSF unfrozen/temp frozen
    - SSF unfrozen/temp unfrozen
    - SSF frozen/temp frozen

    it also stores information about gpi, lat, lon
    and number of total observations

    """

    def __init__(
        self, frozen_flag=2, other_name="k1", metadata_template=None
    ):
        super(FTMetrics, self).__init__(
            other_name=other_name, metadata_template=metadata_template
        )

        self.frozen_flag_value = frozen_flag
        self.result_template.update(
            {
                "ssf_fr_temp_un": np.float32([np.nan]),
                "ssf_fr_temp_fr": np.float32([np.nan]),
                "ssf_un_temp_fr": np.float32([np.nan]),
                "ssf_un_temp_un": np.float32([np.nan]),
                "n_obs": np.int32([0]),
            }
        )

    def calc_metrics(self, data, gpi_info):
        """
        calculates the desired statistics

        Parameters
        ----------
        data : pandas.DataFrame
            with 2 columns, the first column is the reference dataset
            named 'ref'
            the second column the dataset to compare against named 'other'
        gpi_info : tuple
            of (gpi, lon, lat)

        Notes
        -----
        Kendall tau is not calculated at the moment
        because the scipy implementation is very slow which is problematic for
        global comparisons

        """
        dataset = super(FTMetrics, self).calc_metrics(data, gpi_info)

        if len(data) < self.min_obs:
            dataset["status"][0] = eh.INSUFFICIENT_DATA
            return dataset

        ssf, temp = data["ref"].values, data[self.other_name].values
        # SSF <= 1 unfrozen
        # SSF >= 2 frozen

        ssf_frozen = np.where(ssf == self.frozen_flag_value)[0]
        ssf_unfrozen = np.where(ssf != self.frozen_flag_value)[0]

        temp_ssf_frozen = temp[ssf_frozen]
        temp_ssf_unfrozen = temp[ssf_unfrozen]

        # correct classifications
        ssf_temp_frozen = np.where(temp_ssf_frozen < 0)[0]
        ssf_temp_unfrozen = np.where(temp_ssf_unfrozen >= 0)[0]

        # incorrect classifications
        ssf_fr_temp_unfrozen = np.where(temp_ssf_frozen >= 0)[0]
        ssf_un_temp_frozen = np.where(temp_ssf_unfrozen < 0)[0]

        dataset["ssf_fr_temp_un"][0] = len(ssf_fr_temp_unfrozen)
        dataset["ssf_fr_temp_fr"][0] = len(ssf_temp_frozen)

        dataset["ssf_un_temp_fr"][0] = len(ssf_un_temp_frozen)
        dataset["ssf_un_temp_un"][0] = len(ssf_temp_unfrozen)

        dataset["n_obs"][0] = len(data)

        dataset["status"][0] = eh.OK
        return dataset


class HSAF_Metrics(MetadataMetrics):
    """
    This class computes metrics as defined by the H-SAF consortium in
    order to prove the operational readiness of a product. It also stores
    information about gpi, lat, lon and number of observations.
    """

    def __init__(
        self,
        other_name1="k1",
        other_name2="k2",
        dataset_names=None,
        metadata_template=None,
    ):

        super(HSAF_Metrics, self).__init__(
            other_name=other_name1, metadata_template=metadata_template
        )

        # prepare validation dataset names as provided
        self.other_name1 = other_name1
        self.other_name2 = other_name2
        self.df_columns = ["ref", self.other_name1, self.other_name2]
        if dataset_names is None:
            self.ds_names = self.df_columns
        else:
            self.ds_names = dataset_names

        # create lut between df columns and dataset names
        self.ds_names_lut = {}
        for name, col in zip(self.ds_names, self.df_columns):
            self.ds_names_lut[col] = name

        self.tds_names = []
        for combi in itertools.combinations(self.df_columns, 2):
            self.tds_names.append("{:}_and_{:}".format(*combi))

        metrics_common = {"n_obs": np.int32([0])}

        metrics_sds = {
            "snr": np.float32([np.nan]),
            "err_var": np.float32([np.nan]),
            "beta": np.float32([np.nan]),
        }

        metrics_tds = {
            "R": np.float32([np.nan]),
            "p_R": np.float32([np.nan]),
            "rho": np.float32([np.nan]),
            "p_rho": np.float32([np.nan]),
            "bias": np.float32([np.nan]),
            "ubrmsd": np.float32([np.nan]),
        }

        self.seasons = ["ALL", "DJF", "MAM", "JJA", "SON"]

        for season in self.seasons:
            # get template for common metric
            for metric in metrics_common.keys():
                key = "{:}_{:}".format(season, metric)
                self.result_template[key] = metrics_common[metric].copy()

            # get template for single-dataset metric
            for name in self.ds_names:
                for metric in metrics_sds.keys():
                    key = "{:}_{:}_{:}".format(name, season, metric)
                    self.result_template[key] = metrics_sds[metric].copy()

            # get template for two-dataset metric
            for tds_name in self.tds_names:
                split_tds_name = tds_name.split("_and_")
                tds_name_key = "{:}_{:}".format(
                    self.ds_names_lut[split_tds_name[0]],
                    self.ds_names_lut[split_tds_name[1]],
                )
                for metric in metrics_tds.keys():
                    key = "{:}_{:}_{:}".format(tds_name_key, season, metric)
                    self.result_template[key] = metrics_tds[metric].copy()

        self.month_to_season = np.array(
            [
                "",
                "DJF",
                "DJF",
                "MAM",
                "MAM",
                "MAM",
                "JJA",
                "JJA",
                "JJA",
                "SON",
                "SON",
                "SON",
                "DJF",
            ]
        )

    def calc_metrics(self, data, gpi_info):
        """
        calculates the desired statistics

        Parameters
        ----------
        data : pandas.DataFrame
            with 3 columns, the first column is the reference dataset
            named 'ref'
            the second and third column are the datasets to compare against
            named 'k1 and k2'
        gpi_info : tuple
            Grid point info (i.e. gpi, lon, lat)
        """
        dataset = super(HSAF_Metrics, self).calc_metrics(data, gpi_info)

        for season in self.seasons:

            if season != "ALL":
                subset = self.month_to_season[data.index.month] == season
            else:
                subset = np.ones(len(data), dtype=bool)

            # number of observations
            n_obs = subset.sum()
            if n_obs < self.min_obs:
                continue
            dataset["{:}_n_obs".format(season)][0] = n_obs

            # get single dataset metrics
            # calculate SNR
            x = data[self.df_columns[0]].values[subset]
            y = data[self.df_columns[1]].values[subset]
            z = data[self.df_columns[2]].values[subset]

            snr, err, beta = metrics.tcol_metrics(x, y, z)

            for i, name in enumerate(self.ds_names):
                dataset["{:}_{:}_snr".format(name, season)][0] = snr[i]
                dataset["{:}_{:}_err_var".format(name, season)][0] = err[i]
                dataset["{:}_{:}_beta".format(name, season)][0] = beta[i]

            # calculate Pearson correlation
            pearson_R, pearson_p = df_metrics.pearsonr(data)
            pearson_R = pearson_R._asdict()
            pearson_p = pearson_p._asdict()

            # calculate Spearman correlation
            spea_rho, spea_p = df_metrics.spearmanr(data)
            spea_rho = spea_rho._asdict()
            spea_p = spea_p._asdict()

            # scale data to reference in order to calculate absolute metrics
            data_scaled = scale(data, method="min_max")

            # calculate bias
            bias_nT = df_metrics.bias(data_scaled)
            bias_dict = bias_nT._asdict()

            # calculate ubRMSD
            ubRMSD_nT = df_metrics.ubrmsd(data_scaled)
            ubRMSD_dict = ubRMSD_nT._asdict()

            for tds_name in self.tds_names:
                R = pearson_R[tds_name]
                p_R = pearson_p[tds_name]
                rho = spea_rho[tds_name]
                p_rho = spea_p[tds_name]
                bias = bias_dict[tds_name]
                ubRMSD = ubRMSD_dict[tds_name]

                split_tds_name = tds_name.split("_and_")
                tds_name_key = "{:}_{:}".format(
                    self.ds_names_lut[split_tds_name[0]],
                    self.ds_names_lut[split_tds_name[1]],
                )

                dataset["{:}_{:}_R".format(tds_name_key, season)][0] = R
                dataset["{:}_{:}_p_R".format(tds_name_key, season)][0] = p_R
                dataset["{:}_{:}_rho".format(tds_name_key, season)][0] = rho
                dataset["{:}_{:}_p_rho".format(tds_name_key, season)][
                    0
                ] = p_rho
                dataset["{:}_{:}_bias".format(tds_name_key, season)][
                    0
                ] = bias
                dataset["{:}_{:}_ubrmsd".format(tds_name_key, season)][
                    0
                ] = ubRMSD

        dataset["status"][0] = eh.OK
        return dataset


class IntercomparisonMetrics(MetadataMetrics):
    """
    Compare Basic Metrics of multiple satellite data sets to one reference
    data set via:

    - Pearson's R and p
    - Spearman's rho and p
    - RMSD
    - BIAS
    - ubRMSD
    - mse
    - RSS
    - optionally Kendall's tau

    Parameters
    ----------
    refname : str, optional
        Name of the reference column in the DataFrame.
    other_names: tuple, optional (default: ('k1', 'k2', 'k3))
        Name of the column of the non-reference / other datasets in the
        DataFrame.
    calc_rho: boolean, optional
        If True then also Spearman's rho is calculated. This is set to True by
        default.
    calc_tau: boolean, optional
        if True then also tau is calculated. This is set to False by default
        since the calculation of Kendalls tau is rather slow and can
        significantly impact performance of e.g. global validation studies
    metrics_between_nonref : bool, optional (default: False)
        Allow 2-dataset combinations where the ref is not included.
        Warning: can lead to many combinations.
    dataset_names : list, optional (default: None)
        Names of the original datasets, that are used to find the lookup table
        for the df cols.
    metadata_template: dict, optional (default: None)
        See MetadataMetrics

    """

    def __init__(
        self,
        refname="ref",
        other_names=("k1", "k2", "k3"),
        calc_tau=False,
        metrics_between_nonref=False,
        calc_rho=True,
        dataset_names=None,
        metadata_template=None,
    ):

        other_names = list(other_names)
        super(IntercomparisonMetrics, self).__init__(
            other_name=other_names, metadata_template=metadata_template
        )

        # string that splits the dataset names and metric names in the output
        # e.g. 'metric_between_dataset1_and_dataset2'
        self.ds_names_split, self.metric_ds_split = "_and_", "_between_"

        self.df_columns = [refname] + self.other_name

        self.calc_rho = calc_rho
        self.calc_tau = calc_tau

        if dataset_names is None:
            self.ds_names = self.df_columns
        else:
            self.ds_names = dataset_names

        self.ds_names_lut = {}
        for name, col in zip(self.ds_names, self.df_columns):
            self.ds_names_lut[col] = name

        combis = n_combinations(
            self.df_columns,
            2,
            must_include=refname if not metrics_between_nonref else None,
        )

        self.tds_names = []
        for combi in combis:
            self.tds_names.append(
                "{1}{0}{2}".format(self.ds_names_split, *combi)
            )

        # metrics that are equal for all datasets
        metrics_common = ["n_obs"]
        # metrics that are calculated between dataset pairs
        metrics_tds = [
            "R",
            "p_R",
            "rho",
            "p_rho",
            "BIAS",
            "RMSD",
            "mse",
            "RSS",
            "mse_corr",
            "mse_bias",
            "urmsd",
            "mse_var",
            "tau",
            "p_tau",
        ]

        metrics_common = _get_metric_template(metrics_common)
        metrics_tds = _get_metric_template(metrics_tds)

        for metric in metrics_common.keys():
            self.result_template[metric] = metrics_common[metric].copy()

        for tds_name in self.tds_names:
            split_tds_name = tds_name.split(self.ds_names_split)
            tds_name_key = self.ds_names_split.join(
                [
                    self.ds_names_lut[split_tds_name[0]],
                    self.ds_names_lut[split_tds_name[1]],
                ]
            )
            for metric in metrics_tds.keys():
                key = self.metric_ds_split.join([metric, tds_name_key])
                self.result_template[key] = metrics_tds[metric].copy()

        if not calc_rho:
            self.result_template.pop("rho", None)
            self.result_template.pop("p_rho", None)
        if not calc_tau:
            self.result_template.pop("tau", None)
            self.result_template.pop("p_tau", None)

    def calc_metrics(self, data, gpi_info):
        """
        calculates the desired statistics

        Parameters
        ----------
        data : pd.DataFrame
            with >2 columns, the first column is the reference dataset
            named 'ref' other columns are the datasets to compare against
            named 'other_i'
        gpi_info : tuple
            of (gpi, lon, lat)

        Notes
        -----
        Kendall tau is calculation is optional at the moment
        because the scipy implementation is very slow which is problematic for
        global comparisons
        """

        dataset = super(IntercomparisonMetrics, self).calc_metrics(
            data, gpi_info
        )

        subset = np.ones(len(data), dtype=bool)

        n_obs = subset.sum()
        dataset["n_obs"][0] = n_obs
        if n_obs < self.min_obs:
            dataset["status"][0] = eh.INSUFFICIENT_DATA
            return dataset


        # make sure we have the correct order
        data = data[self.df_columns]

        # calculate Pearson correlation
        pearson_R, pearson_p = df_metrics.pearsonr(data)
        pearson_R, pearson_p = pearson_R._asdict(), pearson_p._asdict()

        # calculate bias
        bias_nT = df_metrics.bias(data)
        bias_dict = bias_nT._asdict()

        # calculate RMSD
        rmsd = df_metrics.rmsd(data)
        rmsd_dict = rmsd._asdict()

        # calculate MSE
        mse, mse_corr, mse_bias, mse_var = df_metrics.mse_decomposition(data)
        mse_dict, mse_corr_dict, mse_bias_dict, mse_var_dict = (
            mse._asdict(),
            mse_corr._asdict(),
            mse_bias._asdict(),
            mse_var._asdict(),
        )

        # calculate RSS
        rss = df_metrics.RSS(data)
        rss_dict = rss._asdict()

        # calculate Spearman correlation
        if self.calc_rho:
            rho, p_rho = df_metrics.spearmanr(data)
            rho_dict, p_rho_dict = rho._asdict(), p_rho._asdict()
        else:
            rho = p_rho = p_rho_dict = rho_dict = None

        # calulcate tau
        if self.calc_tau:
            tau, p_tau = df_metrics.kendalltau(data)
            tau_dict, p_tau_dict = tau._asdict(), p_tau._asdict()
        else:
            tau = p_tau = p_tau_dict = tau_dict = None

        # No extra scaling is performed here.
        # always scale for ubRMSD with mean std
        # calculate ubRMSD
        data_scaled = scale(data, method="mean_std")
        ubRMSD_nT = df_metrics.ubrmsd(data_scaled)
        ubRMSD_dict = ubRMSD_nT._asdict()

        for tds_name in self.tds_names:
            R, p_R = pearson_R[tds_name], pearson_p[tds_name]
            bias = bias_dict[tds_name]
            mse = mse_dict[tds_name]
            mse_corr = mse_corr_dict[tds_name]
            mse_bias = mse_bias_dict[tds_name]
            mse_var = mse_var_dict[tds_name]
            rmsd = rmsd_dict[tds_name]
            ubRMSD = ubRMSD_dict[tds_name]
            rss = rss_dict[tds_name]

            if rho_dict and p_rho_dict:
                rho = rho_dict[tds_name]
                p_rho = p_rho_dict[tds_name]
            if tau_dict and p_tau_dict:
                tau = tau_dict[tds_name]
                p_tau = p_tau_dict[tds_name]

            split_tds_name = tds_name.split(self.ds_names_split)
            tds_name_key = self.ds_names_split.join(
                [
                    self.ds_names_lut[split_tds_name[0]],
                    self.ds_names_lut[split_tds_name[1]],
                ]
            )

            dataset[self.metric_ds_split.join(["R", tds_name_key])][0] = R
            dataset[self.metric_ds_split.join(["p_R", tds_name_key])][
                0
            ] = p_R
            dataset[self.metric_ds_split.join(["BIAS", tds_name_key])][
                0
            ] = bias
            dataset[self.metric_ds_split.join(["mse", tds_name_key])][
                0
            ] = mse
            dataset[self.metric_ds_split.join(["mse_corr", tds_name_key])][
                0
            ] = mse_corr
            dataset[self.metric_ds_split.join(["mse_bias", tds_name_key])][
                0
            ] = mse_bias
            dataset[self.metric_ds_split.join(["mse_var", tds_name_key])][
                0
            ] = mse_var
            dataset[self.metric_ds_split.join(["RMSD", tds_name_key])][
                0
            ] = rmsd
            dataset[self.metric_ds_split.join(["urmsd", tds_name_key])][
                0
            ] = ubRMSD
            dataset[self.metric_ds_split.join(["RSS", tds_name_key])][
                0
            ] = rss

            if self.calc_rho:
                dataset[self.metric_ds_split.join(["rho", tds_name_key])][
                    0
                ] = rho
                dataset[self.metric_ds_split.join(["p_rho", tds_name_key])][
                    0
                ] = p_rho
            if self.calc_tau:
                dataset[self.metric_ds_split.join(["tau", tds_name_key])][
                    0
                ] = tau
                dataset[self.metric_ds_split.join(["p_tau", tds_name_key])][
                    0
                ] = p_tau

        dataset["status"][0] = eh.OK
        return dataset


class TCMetrics(MetadataMetrics):
    """
    This class computes triple collocation metrics as defined in the QA4SM
    project. It uses 2 satellite and 1 reference data sets as inputs only.
    It can be extended to perform intercomparison between possible triples
    of more than 3 datasets.
    """

    def __init__(
        self,
        other_names=("k1", "k2"),
        calc_tau=False,
        dataset_names=None,
        tc_metrics_for_ref=True,
        metrics_between_nonref=False,
        metadata_template=None,
    ):
        """
        Triple Collocation metrics as implemented in the QA4SM project.

        Parameters
        ----------
        other_names : tuple, optional (default: ('k1', 'k2'))
            Names of the data sets that are not the reference in the
            data frame.
        calc_tau : bool, optional (default: False)
            Calculate Kendall's Tau (slow)
        dataset_names : tuple, optional (default: None)
            List that maps the names of the satellite dataset columns to their
            real name that will be used in the results file.
        tc_metrics_for_ref : bool, optional (default: False)
            Store TC metrics for the reference data set as well.
        metrics_between_nonref : bool, optional (default: False)
            Allow 2-dataset combinations where the ref is not included.
            Warning: can lead to many combinations.
        metadata_template: dictionary, optional
            A dictionary containing additional fields (and types) of the form
            dict = {'field': np.float32([np.nan]}. Allows users to specify
            information in the job tuple,
            i.e. jobs.append(
                (idx,
                 metadata['longitude'],
                 metadata['latitude'],
                 metadata_dict))
            which is then propagated to the end netCDF results file.

        """
        self.ref_name = "ref"
        other_names = list(other_names)
        super(TCMetrics, self).__init__(
            other_name=other_names, metadata_template=metadata_template
        )

        # string that splits the dataset names and metric names in the output
        # e.g. 'metric_between_dataset1_and_dataset2'
        self.ds_names_split, self.metric_ds_split = "_and_", "_between_"

        self.calc_tau = calc_tau
        self.df_columns = [self.ref_name] + self.other_name

        if dataset_names is None:
            self.ds_names = self.df_columns
        else:
            self.ds_names = dataset_names

        self.ds_names_lut = {}
        for name, col in zip(self.ds_names, self.df_columns):
            self.ds_names_lut[col] = name

        self.metrics_between_nonref = metrics_between_nonref
        self.tds_names, self.thds_names = self._make_names()

        # metrics that are equal for all datasets
        metrics_common = ["n_obs"]
        # metrics that are calculated between dataset pairs
        metrics_tds = [
            "R",
            "p_R",
            "rho",
            "p_rho",
            "BIAS",
            "RMSD",
            "mse",
            "RSS",
            "mse_corr",
            "mse_bias",
            "urmsd",
            "mse_var",
            "tau",
            "p_tau",
        ]
        # metrics that are calculated between dataset triples
        metrics_thds = ["snr", "err_std", "beta"]

        metrics_common = _get_metric_template(metrics_common)
        metrics_tds = _get_metric_template(metrics_tds)

        ignore_ds = [self.ref_name] if not tc_metrics_for_ref else ()
        metrics_thds = _get_tc_metric_template(
            metrics_thds,
            [
                self.ds_names_lut[n]
                for n in self.df_columns
                if n not in ignore_ds
            ],
        )

        for metric in metrics_common.keys():
            self.result_template[metric] = metrics_common[metric].copy()

        for tds_name in self.tds_names:
            split_tds_name = tds_name.split(self.ds_names_split)
            tds_name_key = self.ds_names_split.join(
                [
                    self.ds_names_lut[split_tds_name[0]],
                    self.ds_names_lut[split_tds_name[1]],
                ]
            )
            for metric in metrics_tds.keys():
                key = self.metric_ds_split.join([metric, tds_name_key])
                self.result_template[key] = metrics_tds[metric].copy()

        for thds_name in self.thds_names:
            split_tds_name = thds_name.split(self.ds_names_split)
            thds_name_key = self.ds_names_split.join(
                [
                    self.ds_names_lut[split_tds_name[0]],
                    self.ds_names_lut[split_tds_name[1]],
                    self.ds_names_lut[split_tds_name[2]],
                ]
            )
            for metric, ds in metrics_thds.keys():
                if not any(
                    [
                        self.ds_names_lut[other_ds] == ds
                        for other_ds in thds_name.split(self.ds_names_split)
                    ]
                ):
                    continue
                full_name = "_".join([metric, ds])
                key = self.metric_ds_split.join([full_name, thds_name_key])
                self.result_template[key] = metrics_thds[(metric, ds)].copy()

        if not calc_tau:
            self.result_template.pop("tau", None)
            self.result_template.pop("p_tau", None)

    def _make_names(self):
        tds_names, thds_names = [], []
        combis_2 = n_combinations(
            self.df_columns,
            2,
            must_include=[self.ref_name]
            if not self.metrics_between_nonref
            else None,
        )
        combis_3 = n_combinations(
            self.df_columns, 3, must_include=[self.ref_name]
        )

        for combi in combis_2:
            tds_names.append(self.ds_names_split.join(combi))

        for combi in combis_3:
            thds_names.append(
                "{1}{0}{2}{0}{3}".format(self.ds_names_split, *combi)
            )

        return tds_names, thds_names

    def _tc_res_dict(self, res):
        """name is the TC metric name and res the according named tuple"""
        res_dict = {}

        metric = np.array([type(r).__name__ for r in res])
        assert all(metric == np.repeat(metric[0], metric.size))

        for r in res:
            r_d = r._asdict()
            ds = self.ds_names_split.join(list(r_d.keys()))
            res_dict[ds] = dict(zip(list(r_d.keys()), list(r_d.values())))

        return res_dict

    def calc_metrics(self, data, gpi_info):
        """
        Calculate Triple Collocation metrics

        Parameters
        ----------
        data : pd.DataFrame
            with >2 columns, the first column is the reference dataset
            named 'ref' other columns are the data sets to compare against
            named 'other_i'
        gpi_info : tuple
            of (gpi, lon, lat)

        Notes
        -----
        Kendall tau is calculation is optional at the moment
        because the scipy implementation is very slow which is problematic for
        global comparisons
        """

        dataset = copy.deepcopy(self.result_template)

        dataset["gpi"][0] = gpi_info[0]
        dataset["lon"][0] = gpi_info[1]
        dataset["lat"][0] = gpi_info[2]

        if self.metadata_template is not None:
            for key, value in self.metadata_template.items():
                dataset[key][0] = gpi_info[3][key]

        # number of observations
        subset = np.ones(len(data), dtype=bool)

        n_obs = subset.sum()
        dataset["n_obs"][0] = n_obs
        if n_obs < self.min_obs:
            dataset["status"][0] = eh.INSUFFICIENT_DATA
            return dataset


        # calculate Pearson correlation
        pearson_R, pearson_p = df_metrics.pearsonr(data)
        pearson_R, pearson_p = pearson_R._asdict(), pearson_p._asdict()
        # calculate Spearman correlation
        spea_rho, spea_p = df_metrics.spearmanr(data)
        spea_rho, spea_p = spea_rho._asdict(), spea_p._asdict()
        # calculate bias
        bias_nT = df_metrics.bias(data)
        bias_dict = bias_nT._asdict()
        # calculate RMSD
        rmsd = df_metrics.rmsd(data)
        rmsd_dict = rmsd._asdict()
        # calculate MSE
        mse, mse_corr, mse_bias, mse_var = df_metrics.mse_decomposition(data)
        mse_dict = mse._asdict()
        mse_corr_dict = mse_corr._asdict()
        mse_bias_dict = mse_bias._asdict()
        mse_var_dict = mse_var._asdict()
        # calculate RSS
        rss = df_metrics.RSS(data)
        rss_dict = rss._asdict()
        # calculate ubRMSD
        # todo: we could use the TC derived scaling parameters here?
        data_scaled = scale(data, method="mean_std")
        ubRMSD_nT = df_metrics.ubrmsd(data_scaled)
        ubRMSD_dict = ubRMSD_nT._asdict()
        # calulcate tau
        if self.calc_tau:
            tau, p_tau = df_metrics.kendalltau(data)
            tau_dict, p_tau_dict = tau._asdict(), p_tau._asdict()
        else:
            tau = p_tau = p_tau_dict = tau_dict = None
        # calculate TC metrics
        ref_ind = np.where(np.array(data.columns) == self.ref_name)[0][0]
        snrs, err_stds, betas = df_metrics.tcol_metrics(
            data, ref_ind=ref_ind
        )
        snr_dict = self._tc_res_dict(snrs)
        err_std_dict = self._tc_res_dict(err_stds)
        beta_dict = self._tc_res_dict(betas)

        # store TC results
        for thds_name in self.thds_names:
            snr = snr_dict[thds_name]
            err_std = err_std_dict[thds_name]
            beta = beta_dict[thds_name]

            split_thds_name = thds_name.split(self.ds_names_split)
            thds_name_key = self.ds_names_split.join(
                [
                    self.ds_names_lut[split_thds_name[0]],
                    self.ds_names_lut[split_thds_name[1]],
                    self.ds_names_lut[split_thds_name[2]],
                ]
            )

            for metr, res in dict(
                snr=snr, err_std=err_std, beta=beta
            ).items():
                for ds, ds_res in res.items():
                    m_ds = "{}_{}".format(metr, self.ds_names_lut[ds])
                    n = "{}{}{}".format(
                        m_ds, self.metric_ds_split, thds_name_key
                    )
                    if n in dataset.keys():
                        dataset[n][0] = ds_res

        # Store basic metrics results
        for tds_name in self.tds_names:
            R, p_R = pearson_R[tds_name], pearson_p[tds_name]
            rho, p_rho = spea_rho[tds_name], spea_p[tds_name]
            bias = bias_dict[tds_name]
            mse = mse_dict[tds_name]
            mse_corr = mse_corr_dict[tds_name]
            mse_bias = mse_bias_dict[tds_name]
            mse_var = mse_var_dict[tds_name]
            rmsd = rmsd_dict[tds_name]
            ubRMSD = ubRMSD_dict[tds_name]
            rss = rss_dict[tds_name]

            if tau_dict and p_tau_dict:
                tau = tau_dict[tds_name]
                p_tau = p_tau_dict[tds_name]

            split_tds_name = tds_name.split(self.ds_names_split)
            tds_name_key = self.ds_names_split.join(
                [
                    self.ds_names_lut[split_tds_name[0]],
                    self.ds_names_lut[split_tds_name[1]],
                ]
            )

            dataset[self.metric_ds_split.join(["R", tds_name_key])][0] = R
            dataset[self.metric_ds_split.join(["p_R", tds_name_key])][
                0
            ] = p_R
            dataset[self.metric_ds_split.join(["rho", tds_name_key])][
                0
            ] = rho
            dataset[self.metric_ds_split.join(["p_rho", tds_name_key])][
                0
            ] = p_rho
            dataset[self.metric_ds_split.join(["BIAS", tds_name_key])][
                0
            ] = bias
            dataset[self.metric_ds_split.join(["mse", tds_name_key])][
                0
            ] = mse
            dataset[self.metric_ds_split.join(["mse_corr", tds_name_key])][
                0
            ] = mse_corr
            dataset[self.metric_ds_split.join(["mse_bias", tds_name_key])][
                0
            ] = mse_bias
            dataset[self.metric_ds_split.join(["mse_var", tds_name_key])][
                0
            ] = mse_var
            dataset[self.metric_ds_split.join(["RMSD", tds_name_key])][
                0
            ] = rmsd
            dataset[self.metric_ds_split.join(["urmsd", tds_name_key])][
                0
            ] = ubRMSD
            dataset[self.metric_ds_split.join(["RSS", tds_name_key])][
                0
            ] = rss

            if self.calc_tau:
                dataset[self.metric_ds_split.join(["tau", tds_name_key])][
                    0
                ] = tau
                dataset[self.metric_ds_split.join(["p_tau", tds_name_key])][
                    0
                ] = p_tau

        dataset["status"][0] = eh.OK
        return dataset


class RollingMetrics(MetadataMetrics):
    """
    This class computes rolling metrics for Pearson R and RMSD. It also stores
    information about gpi, lat, lon and number of observations.

    Parameters
    ----------
    other_name: string or tuple, optional
        Name of the column of the non-reference / other dataset in the
        pandas DataFrame
    metadata_template: dictionary, optional
        A dictionary containing additional fields (and types) of the form
        ``dict = {'field': np.float32([np.nan]}``. Allows users to specify
        information in the job tuple, i.e.::

            jobs.append((idx, metadata['longitude'], metadata['latitude'],
                         metadata_dict))``

        which is then propagated to the end netCDF results file.
    """

    def __init__(self, other_name="k1", metadata_template=None):

        super(RollingMetrics, self).__init__(
            other_name=other_name, metadata_template=metadata_template
        )

        self.basic_metrics = ["R", "p_R", "RMSD"]
        self.result_template.update(_get_metric_template(self.basic_metrics))

    def calc_metrics(
        self, data, gpi_info, window_size="30d", center=True, min_periods=2
    ):
        """
        Calculate the desired statistics.

        Parameters
        ----------
        data : pandas.DataFrame
            with 2 columns, the first column is the reference dataset
            named 'ref' the second column the dataset to compare
            against named 'other'
        gpi_info : tuple
            of (gpi, lon, lat)
        window_size : string
            Window size defined as string.
        center : bool, optional
            Set window at the center.
        min_periods : int, optional
            Minimum number of observations in window required for computation.
        """
        dataset = super(RollingMetrics, self).calc_metrics(data, gpi_info)

        if len(data) < self.min_obs:
            dataset["status"][0] = eh.INSUFFICIENT_DATA
            return dataset

        xy = data.to_numpy()
        timestamps = data.index.to_julian_date().values
        window_size_jd = pd.Timedelta(
            window_size
        ).to_numpy() / np.timedelta64(1, "D")
        pr_arr, rmsd_arr = metrics.rolling_pr_rmsd(
            timestamps,
            xy[:, 0],
            xy[:, 1],
            window_size_jd,
            center,
            min_periods,
        )

        dataset["time"] = np.array([data.index])
        dataset["R"] = np.array([pr_arr[:, 0]])
        dataset["p_R"] = np.array([pr_arr[:, 1]])
        dataset["RMSD"] = np.array([rmsd_arr[:]])

        dataset["status"][0] = eh.OK
        return dataset


def get_dataset_names(ref_key, datasets, n=3):
    """
    Get dataset names in correct order as used in the validation framework

    - reference dataset = ref
    - first other dataset = k1
    - second other dataset = k2

    This is important to correctly iterate through the H-SAF metrics and to
    save each metric with the name of the used datasets

    Parameters
    ----------
    ref_key: basestring
        Name of the reference dataset
    datasets: dict
        Dictionary of dictionaries as provided to the validation framework
        in order to perform the validation process.

    Returns
    -------
    dataset_names: list
        List of the dataset names in correct order

    """
    ds_dict = {}
    for ds in datasets.keys():
        ds_dict[ds] = datasets[ds]["columns"]
    ds_names = get_result_names(ds_dict, ref_key, n)
    dataset_names = []
    for name in ds_names[0]:
        dataset_names.append(name[0])

    return dataset_names


class PairwiseMetricsMixin:
    def _pairwise_metric_names(self):
        """
        Returns a list of metric names to be calculated between pairs.
        """
        metrics = [
            "R",
            "p_R",
            "BIAS",
            "RMSD",
            "mse",
            "RSS",
            "mse_corr",
            "mse_bias",
            "urmsd",
            "mse_var",
        ]
        if self.calc_spearman:
            metrics += ["rho", "p_rho"]
        if self.calc_kendall:
            metrics += ["tau", "p_tau"]
        if self.analytical_cis:
            metrics += [
                "BIAS_ci_lower", "BIAS_ci_upper",
                "urmsd_ci_lower", "urmsd_ci_upper",
                "R_ci_lower", "R_ci_upper",
            ]
            if self.calc_spearman:
                metrics += ["rho_ci_lower", "rho_ci_upper"]
            if self.calc_kendall:
                metrics += ["tau_ci_lower", "tau_ci_upper"]
        if self.bootstrap_cis:
            metrics += ["mse_var_ci_lower", "mse_var_ci_upper",
                        "mse_corr_ci_lower", "mse_corr_ci_upper",
                        "mse_bias_ci_lower", "mse_bias_ci_upper",
                        "RMSD_ci_lower", "RMSD_ci_upper",
                        "mse_ci_lower", "mse_ci_upper",
                        "RSS_ci_lower", "RSS_ci_upper",
                        ]
        return metrics

    def _calc_pairwise_metrics(
        self, x, y, mx, my, varx, vary, cov, result, suffix=""
    ):
        """
        Calculates pairwise metrics, making use of pre-computed moments.

        Parameters
        ----------
        x, y : np.ndarray
            Data as numpy matrix
        mx, my : float
            Means
        varx, vary : float
            Variances
        cov : float
            Covariance
        result : dict
            Result template dictionary. This is where the results are written
            to.
        suffix : str, optional
            Suffix to be used for storing the metric calculation result. E.g.,
            with ``suffix=_between_k1_and_k2``, bias will be stored as
            ``BIAS_between_k1_and_k2``.
        """
        n_obs = len(x)
        result["BIAS" + suffix][0] = mx - my

        mse_corr = _mse_corr_from_moments(mx, my, varx, vary, cov)
        mse_var = _mse_var_from_moments(mx, my, varx, vary, cov)
        mse_bias = _mse_bias_from_moments(mx, my, varx, vary, cov)
        mse = max(mse_corr + mse_var + mse_bias, 0)
        result["mse_corr" + suffix][0] = mse_corr
        result["mse_var" + suffix][0] = mse_var
        result["mse_bias" + suffix][0] = mse_bias
        result["mse" + suffix][0] = mse

        result["RSS" + suffix][0] = mse * n_obs
        result["RMSD" + suffix][0] = np.sqrt(mse)
        result["urmsd" + suffix][0] = np.sqrt(mse - mse_bias)

        R, p_R = _pearsonr_from_moments(varx, vary, cov, n_obs)
        result["R" + suffix][0] = R
        result["p_R" + suffix][0] = p_R

        if self.calc_spearman:
            (
                result["rho" + suffix][0],
                result["p_rho" + suffix][0],
            ) = stats.spearmanr(x, y)
        if self.calc_kendall:
            (
                result["tau" + suffix][0],
                result["p_tau" + suffix][0],
            ) = stats.kendalltau(x, y)

        if self.analytical_cis:
            (
                result["BIAS_ci_lower" + suffix][0],
                result["BIAS_ci_upper" + suffix][0],
            ) = _bias_ci_from_moments(0.05, mx, my, varx, vary, cov, n_obs)

            (
                result["urmsd_ci_lower" + suffix][0],
                result["urmsd_ci_upper" + suffix][0],
            ) = ubrmsd_ci(x, y, result["urmsd" + suffix][0])

            (
                result["R_ci_lower" + suffix][0],
                result["R_ci_upper" + suffix][0],
            ) = pearson_r_ci(x, y, result["R" + suffix][0])

            if self.calc_spearman:
                (
                    result["rho_ci_lower" + suffix][0],
                    result["rho_ci_upper" + suffix][0],
                ) = spearman_r_ci(x, y, result["rho" + suffix][0])
            if self.calc_kendall:
                (
                    result["tau_ci_lower" + suffix][0],
                    result["tau_ci_upper" + suffix][0],
                ) = kendall_tau_ci(x, y, result["tau" + suffix][0])

        if self.bootstrap_cis:
            for m in ["mse_var", "mse_corr", "mse_bias", "mse", "RMSD"]:
                if m == "mse":
                    metric_func = pairwise.msd
                elif m == "RMSD":
                    metric_func = pairwise.rmsd
                else:
                    metric_func = getattr(pairwise, m)
                kwargs = {}
                if hasattr(self, 'bootstrap_alpha'):
                    kwargs['alpha'] = getattr(self, 'bootstrap_alpha')
                if hasattr(self, 'bootstrap_min_obs'):
                    kwargs['minimum_data_length'] = getattr(
                        self, 'bootstrap_min_obs')
                _, lb, ub = with_bootstrapped_ci(metric_func, x, y, **kwargs)
                result[f"{m}_ci_lower" + suffix][0] = lb
                result[f"{m}_ci_upper" + suffix][0] = ub


class PairwiseIntercomparisonMetrics(MetadataMetrics, PairwiseMetricsMixin):
    """
    Basic metrics for comparison of two datasets:

    - RMSD
    - BIAS
    - ubRMSD
    - mse and decomposition (mse_var, mse_corr, mse_bias)
    - RSS
    - Pearson's R and p
    - Spearman's rho and p (optional)
    - Kendall's tau and p (optional)

    Additionally, confidence intervals for these metrics can be calculated
    (optional).

    **NOTE**: When using this within a
      ``pytesmo.validation_framework.validation.Validation``, use
      ``temporal_matcher=make_combined_temporal_matcher(<window>)`` as keyword
      argument. ``make_combined_temporal_matcher`` can be imported from
      ``pytesmo.validation_framework.temporal_matchers``.

    Parameters
    ----------
    min_obs : int, optional
        Minimum number of observations required to calculate metrics. Default
        is 10.
    calc_spearman : bool, optional
        Whether to calculate Spearman's rank correlation coefficient. Default
        is True.
    calc_kendall : bool, optional
        Whether to calculate Kendall's rank correlation coefficient. Default is
        True.
    analytical_cis : bool, optional (default: True)
        Whether to calculate analytical confidence intervals for the following
        metrics:
            - BIAS
            - mse_bias
            - RMSD
            - urmsd
            - mse
            - R
            - rho (only if ``calc_spearman=True``)
            - tau (only if ``calc_kendall=True``)
    bootstrap_cis: bool, optional (default: False)
        Whether to calculate bootstrap confidence intervals for the following
        metrics:
            - mse_corr
            - mse_var
        The default is `False`. This might be a lot of computational effort.
    bootstrap_min_obs: int, optional (default: 100)
        Minimum number of observations to draw from the time series for boot-
        strapping.
    bootstrap_alpha: float, optional (default: 0.05)
        Confidence level.
    """

    def __init__(
        self,
        min_obs=10,
        calc_spearman=True,
        calc_kendall=True,
        analytical_cis=True,
        bootstrap_cis=False,
        bootstrap_min_obs=100,
        bootstrap_alpha=0.05,
        metadata_template=None,
    ):
        super().__init__(min_obs=min_obs, metadata_template=metadata_template)

        self.calc_spearman = calc_spearman
        self.calc_kendall = calc_kendall
        self.analytical_cis = analytical_cis
        self.bootstrap_cis = bootstrap_cis
        self.bootstrap_min_obs = bootstrap_min_obs
        self.bootstrap_alpha = bootstrap_alpha

        metrics = self._pairwise_metric_names()
        metrics.append("n_obs")
        self.result_template.update(_get_metric_template(metrics))

    def calc_metrics(self, data, gpi_info):
        """
        Calculates pairwise metrics.

        Parameters
        ----------
        data : pd.DataFrame
            DataFrame with 2 columns between which metrics should be
            calculated.
        gpi_info : tuple
            (gpi, lon, lat)
        """
        result = super().calc_metrics(data, gpi_info)

        n_obs = len(data)
        result["n_obs"][0] = n_obs
        if n_obs < self.min_obs:
            result["status"][0] = eh.INSUFFICIENT_DATA
            warnings.warn(
                "Not enough observations to calculate metrics.", UserWarning
            )
            return result

        data_matrix = data.values
        x = data_matrix[:, 0]
        y = data_matrix[:, 1]

        # we can calculate almost all metrics from moments
        mx, my, varx, vary, cov = _moments_welford(x, y)
        self._calc_pairwise_metrics(x, y, mx, my, varx, vary, cov, result)
        result["status"][0] = eh.OK
        return result


class TripleCollocationMetrics(MetadataMetrics, PairwiseMetricsMixin):
    """
    Computes triple collocation metrics

    The triple collocation metrics calculated are:

    - SNR
    - error standard deviation
    - linear scaling/multiplicative (first-order) bias

    **NOTE**: When using this within a
      ``pytesmo.validation_framework.validation.Validation``, use
      ``temporal_matcher=make_combined_temporal_matcher(<window>)`` as keyword
      argument. ``make_combined_temporal_matcher`` can be imported from
      ``pytesmo.validation_framework.temporal_matchers``.

    Parameters
    ----------
    refname : str
        Name of the reference column that is passed to ``calc_metrics``. This
        will also be used to name the results. **Make sure that you set
        ``rename_cols=False`` in the call to ``Validation.calc``, otherwise the
        names will be wrong.**
    min_obs : int, optional
        Minimum number of observations required to calculate metrics. Default
        is 10.
    bootstrap_cis
        Whether to calculate bootstrap confidence intervals for triple
        collocation metrics.
        The default is `False`. This might be a lot of computational effort.
    bootstrap_min_obs: int, optional (default: 100)
        Minimum number of observations to draw from the time series for boot-
        strapping.
    bootstrap_alpha: float, optional (default: 0.05)
        Confidence level.
    metadata_template: dict, optional (default: None)
        A dictionary containing additional fields (and types) of the form
        dict = {'field': np.float32([np.nan]}. Allows users to specify
        information in the job tuple, i.e. jobs.append((idx,
        metadata['longitude'], metadata['latitude'], metadata_dict)) which
        is then propagated to the end netCDF results file.
    """

    def __init__(
        self,
        refname,
        min_obs=10,
        bootstrap_cis=False,
        bootstrap_min_obs=100,
        bootstrap_alpha=0.05,
        metadata_template=None,
    ):

        super().__init__(min_obs=min_obs, metadata_template=metadata_template)

        self.bootstrap_cis = bootstrap_cis
        self.bootstrap_min_obs = bootstrap_min_obs
        self.bootstrap_alpha = bootstrap_alpha
        self.refname = refname
        self.result_template.update(_get_metric_template(["n_obs"]))

    def _get_metric_template(self, refname, othernames):
        # othernames must have length 2 here!
        result_template = {}
        tcol_metrics = ["snr", "err_std", "beta"]
        tcol_template = _get_tc_metric_template(
            tcol_metrics, (refname, *othernames)
        )
        result_template.update(copy.copy(tcol_template))
        if self.bootstrap_cis:
            for metric in tcol_metrics:
                for ds in [refname] + othernames:
                    result_template[(metric + "_ci_lower", ds)] = np.array(
                        [np.nan], dtype=np.float32
                    )
                    result_template[(metric + "_ci_upper", ds)] = np.array(
                        [np.nan], dtype=np.float32
                    )
        return result_template

    def calc_metrics(self, data, gpi_info):
        """
        Calculates triple collocation metrics.

        Parameters
        ----------
        data : pd.DataFrame
            DataFrame with one reference column and two other columns between
            which the metrics are calculated. The name of the reference column
            must be the same as used in the constructor.
            Make sure to use ``rename_cols`` for ``Validation.calc``, so that
            the names are correct.
        gpi_info : tuple
            (gpi, lon, lat)
        """
        result = super().calc_metrics(data, gpi_info)
        n_obs = len(data)
        result["n_obs"][0] = n_obs

        # get the remaining metrics template for this specific combination
        othernames = list(data.columns)
        othernames.remove(self.refname)
        result.update(self._get_metric_template(self.refname, othernames))

        if n_obs < self.min_obs:
            result["status"][0] = eh.INSUFFICIENT_DATA
            warnings.warn(
                "Not enough observations to calculate metrics.", UserWarning
            )
            return result

        # calculate triple collocation metrics
        ds_names = (self.refname, *othernames)
        arrays = (data[name].values for name in ds_names)
        if not self.bootstrap_cis:
            res = tcol_metrics(*arrays)
            for i, name in enumerate(ds_names):
                for j, metric in enumerate(["snr", "err_std", "beta"]):
                    result[(metric, name)][0] = res[j][i]
        else:
            try:
                # handle failing bootstrapping because e.g.
                # too small sample size
                res = tcol_metrics_with_bootstrapped_ci(
                    *arrays, minimum_data_length=self.bootstrap_min_obs,
                    alpha=self.bootstrap_alpha)
                for i, name in enumerate(ds_names):
                    for j, metric in enumerate(["snr", "err_std", "beta"]):
                        result[(metric, name)][0] = res[j][0][i]
                        result[(metric + "_ci_lower", name)][0] = res[j][1][i]
                        result[(metric + "_ci_upper", name)][0] = res[j][2][i]
                result["status"][0] = eh.OK
            except ValueError:
                # if the calculation fails, the template results (np.nan) are used
                result["status"] = eh.METRICS_CALCULATION_FAILED
        return result



# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Christoph Paulik christoph.paulik@geo.tuwien.ac.at
# Creation date: 2014-06-30


"""
Tests for grouping functions
"""

from datetime import date, datetime
from pytesmo.time_series.grouping import grouped_dates_between


def test_grouped_dates_between():
    tstamps_should = [datetime(2007, 1, 10),
                      datetime(2007, 1, 20),
                      datetime(2007, 1, 31),
                      datetime(2007, 2, 10)]
    tstamps = grouped_dates_between(date(2007, 1, 1), date(2007, 2, 1))
    assert tstamps == tstamps_should


def test_grouped_dates_between_start():
    tstamps_should = [datetime(2007, 1, 1),
                      datetime(2007, 1, 11),
                      datetime(2007, 1, 21),
                      datetime(2007, 2, 1)]
    tstamps = grouped_dates_between(
        date(2007, 1, 1), date(2007, 2, 1), start=True)
    assert tstamps == tstamps_should



# Copyright (c) 2014,Vienna University of Technology, Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology, Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from warnings import warn
from repurpose.resample import resample_to_grid_only_valid_return, resample_to_grid, hamming_window

warn("resampling functions have moved to the repurpose package and will be"
     "removed from pytesmo soon!")



# Copyright (c) 2016,Vienna University of Technology,
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice,
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice,
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology, Department of
#     Geodesy and Geoinformation nor the names of its contributors may be used
#     to endorse or promote products derived from this software without specific
#     prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY, DEPARTMENT OF
# GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

'''
Tests for the results managers.
'''

import tempfile
import numpy as np
import os
import netCDF4
import pandas as pd

from pytesmo.validation_framework.results_manager import netcdf_results_manager, PointDataResults


def test_netcdf_result_manager_n2():

    tst_results = {
        (('DS1', 'x'), ('DS3', 'y')): {
            'n_obs': np.array([1000], dtype=np.int32),
            'tau': np.array([np.nan], dtype=np.float32),
            'gpi': np.array([4], dtype=np.int32),
            'RMSD': np.array([0.], dtype=np.float32),
            'lon': np.array([4.]),
            'p_tau': np.array([np.nan], dtype=np.float32),
            'BIAS': np.array([0.], dtype=np.float32),
            'p_rho': np.array([0.], dtype=np.float32),
            'rho': np.array([1.], dtype=np.float32),
            'lat': np.array([4.]),
            'R': np.array([1.], dtype=np.float32),
            'p_R': np.array([0.], dtype=np.float32)},
        (('DS1', 'x'), ('DS2', 'y')): {
            'n_obs': np.array([1000], dtype=np.int32),
            'tau': np.array([np.nan], dtype=np.float32),
            'gpi': np.array([4], dtype=np.int32),
            'RMSD': np.array([0.], dtype=np.float32),
            'lon': np.array([4.]),
            'p_tau': np.array([np.nan], dtype=np.float32),
            'BIAS': np.array([0.], dtype=np.float32),
            'p_rho': np.array([0.], dtype=np.float32),
            'rho': np.array([1.], dtype=np.float32),
            'lat': np.array([4.]),
            'R': np.array([1.], dtype=np.float32),
            'p_R': np.array([0.], dtype=np.float32)},
        (('DS1', 'x'), ('DS3', 'x')): {
            'n_obs': np.array([1000], dtype=np.int32),
            'tau': np.array([np.nan], dtype=np.float32),
            'gpi': np.array([4], dtype=np.int32),
            'RMSD': np.array([0.], dtype=np.float32),
            'lon': np.array([4.]),
            'p_tau': np.array([np.nan], dtype=np.float32),
            'BIAS': np.array([0.], dtype=np.float32),
            'p_rho': np.array([0.], dtype=np.float32),
            'rho': np.array([1.], dtype=np.float32),
            'lat': np.array([4.]),
            'R': np.array([1.], dtype=np.float32),
            'p_R': np.array([0.], dtype=np.float32)}}

    tempdir = tempfile.mkdtemp()
    netcdf_results_manager(tst_results, tempdir)
    assert sorted(os.listdir(tempdir)) == sorted(['DS1.x_with_DS3.x.nc',
                                                  'DS1.x_with_DS3.y.nc',
                                                  'DS1.x_with_DS2.y.nc'])

    # check a few variable in the file
    with netCDF4.Dataset(os.path.join(tempdir, 'DS1.x_with_DS3.x.nc')) as ds:
        assert ds.variables['lon'][:] == np.array([4])
        assert ds.variables['n_obs'][:] == np.array([1000])


def test_netcdf_result_manager_n3():

    tst_results = {
        (('DS1', 'x'), ('DS2', 'y'), ('DS3', 'x')): {
            'n_obs': np.array([1000], dtype=np.int32),
            'tau': np.array([np.nan], dtype=np.float32),
            'gpi': np.array([4], dtype=np.int32),
            'RMSD': np.array([0.], dtype=np.float32),
            'lon': np.array([4.]),
            'p_tau': np.array([np.nan], dtype=np.float32),
            'BIAS': np.array([0.], dtype=np.float32),
            'p_rho': np.array([0.], dtype=np.float32),
            'rho': np.array([1.], dtype=np.float32),
            'lat': np.array([4.]),
            'R': np.array([1.], dtype=np.float32),
            'p_R': np.array([0.], dtype=np.float32)},
        (('DS1', 'x'), ('DS2', 'y'), ('DS3', 'y')): {
            'n_obs': np.array([1000], dtype=np.int32),
            'tau': np.array([np.nan], dtype=np.float32),
            'gpi': np.array([4], dtype=np.int32),
            'RMSD': np.array([0.], dtype=np.float32),
            'lon': np.array([4.]),
            'p_tau': np.array([np.nan], dtype=np.float32),
            'BIAS': np.array([0.], dtype=np.float32),
            'p_rho': np.array([0.], dtype=np.float32),
            'rho': np.array([1.], dtype=np.float32),
            'lat': np.array([4.]),
            'R': np.array([1.], dtype=np.float32),
            'p_R': np.array([0.], dtype=np.float32)}}

    tempdir = tempfile.mkdtemp()
    netcdf_results_manager(tst_results, tempdir)
    assert (
        sorted(os.listdir(tempdir))
        == sorted(['DS1.x_with_DS2.y_with_DS3.x.nc',
                   'DS1.x_with_DS2.y_with_DS3.y.nc'])
    )

    # check a few variable in the file
    with netCDF4.Dataset(
            os.path.join(tempdir, 'DS1.x_with_DS2.y_with_DS3.x.nc')
    ) as ds:
        assert ds.variables['lon'][:] == np.array([4])
        assert ds.variables['n_obs'][:] == np.array([1000])


def test_netcdf_results_manager_ts():
    results = {
        (('a1', 'x1'), ('b1', 'y1')): {
            'time': np.array([
                pd.date_range('2000-01-01', '2000-01-03', freq='D'),
                pd.date_range('2000-05-01', '2000-05-05', freq='D')
            ], dtype="object"),
            'tvar': np.array([
                np.array([1, 2, 3]),
                np.array([1, 2, 3, 4, 5])
            ], dtype="object"),
            'lvar': np.array([99, 100]),
            'lon': np.array([1., 2.]),
            'lat': np.array([1., 2.])
        },
        (('a2', 'x2'), ('b2', 'y2')): {
            'time': np.array([
                pd.date_range('2003-01-01', '2003-01-02', freq='D')]),
            'tvar': np.array([np.array([1, 2])]),
            'lvar': np.array([99]),
            'lon': np.array([1.]),
            'lat': np.array([1.])
        }
    }

    tempdir = tempfile.mkdtemp()
    netcdf_results_manager(
        results=results, save_path=tempdir, ts_vars=['tvar'],
        attr={'tvar': {'long_name': 'Time var'},
              'lvar': {'long_name': 'Loc var'}}
    )

    ds = PointDataResults(
        os.path.join(tempdir, 'a1.x1_with_b1.y1.nc'), read_only=True
    )
    ts = ds.read_ts(0)
    assert ts.loc['2000-01-02', 'tvar'] == 2
    df = ds.read_loc(None)
    assert np.all(df.loc[0, :] == ds.read_loc(0))
    assert df.loc[1, 'lvar'] == 100



"""
Metrics for pairs and triplets
==============================

This module provides functions to calculate metrics and their associated
confidence intervals.

Pairwise Metrics
----------------

For pairwise metrics, confidence intervals can be calculated analytically (if
available) or via boostrapping. Example with bias::

    from pytesmo.metrics import bias, with_analytical_ci

    x, y = np.random.randn(100), np.rand.randn(100)
    # only metric calculation
    b = bias(x, y)

    # with analytical CI
    b, lower, upper = with_analytical_ci(bias, x, y, alpha=0.05)

    # using bootstrapping instead (uses nsamples=1000 by default)
    b, lower, upper = with_bootstrapped_ci(bias, x, y, alpha=0.05)

Analytical CIs are available for

* bias
* rmsd
* nrmsd
* ubrmsd
* mse
* pearson_r
* spearman_r
* kendall_tau

You can use :py:func:`pytesmo.metrics.has_analytical_ci` to check
programmatically whether analytical CIs are available.

Triple Collocation Metrics
--------------------------

Triple collocation metric can be calculated with the function
:py:func:`pytesmo.metrics.tcol_metrics`. For bootstrapping CIs, use
:py:func:`pytesmo.metrics.tcol_metrics_with_bootstrapped_ci`::

    from pytesmo.metrics import tcol_metrics, tcol_metric

    x, y, z = ...

    # only metrics
    snr, err_std, beta = tcol_metrics(x, y, z)

    # with cis, using nsamples=1000 and alpha=0.05 (default)
    snr_and_ci, err_std_and_ci, beta_and_ci = (
        tcol_metrics_with_bootstrapped_ci(x, y, z)
    )
    snr, lower_snr, upper_snr = snr_and_ci



Developer notes
---------------
Pairwise metrics are implemented in :py:mod:`pytesmo.metrics.pairwise`, triple
collocation metrics in :py:mod:`pytesmo.metrics.tcol`.
"""


from pytesmo.metrics.pairwise import (
    aad,
    bias,
    mad,
    msd,
    rmsd,
    nrmsd,
    ubrmsd,
    mse_bias,
    mse_var,
    mse_corr,
    mse_decomposition,
    RSS,
    pearson_r,
    spearman_r,
    kendall_tau,
    nash_sutcliffe,
    index_of_agreement,
    rolling_pr_rmsd,
)
from pytesmo.metrics.confidence_intervals import (
    has_analytical_ci,
    with_analytical_ci,
    with_bootstrapped_ci,
    tcol_metrics_with_bootstrapped_ci,
)

from pytesmo.metrics.tcol import (
    tcol_metrics,
    ecol
)

from pytesmo.metrics.deprecated import (
    mse,
    tcol_error,
    tcol_snr,
    pearsonr,
    spearmanr,
    kendalltau,
    pearson_conf,
    pearsonr_recursive,
)

__all__ = [
    "bias",
    "mse_bias",
    "mse_var",
    "mse_corr",
    "mse_decomposition",
    "RSS",
    "rolling_pr_rmsd",
    "aad",
    "mad",
    "msd",
    "rmsd",
    "nrmsd",
    "ubrmsd",
    "pearson_r",
    "spearman_r",
    "kendall_tau",
    "index_of_agreement",
    "nash_sutcliffe",
    "has_analytical_ci",
    "with_analytical_ci",
    "with_bootstrapped_ci",
    "tcol_metrics_with_bootstrapped_ci",
    "tcol_metrics",
    "ecol",
]



# Copyright (c) 2016, TU Wien
# Department of Geodesy and Geoinformation
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#   * Neither the name of the Vienna University of Technology,
#     Department of Geodesy and Geoinformation nor the
#     names of its contributors may be used to endorse or promote products
#     derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
Tests for the time series filtering module
'''
import pytesmo.time_series.filtering as filtering
import numpy as np
import pandas as pd


def test_moving_average():
    """
    Test moving average filter.
    """
    test_jd = np.arange(10, dtype=np.double)
    test_data = np.array(
        [1, 2, 3, 4, -999999.0, 6, 7, 8, 9, np.nan], dtype=np.double)

    ser = pd.Series(test_data, index=test_jd)

    filtered = filtering.moving_average(ser, window_size=5)

    np.testing.assert_allclose(filtered.values, [2., 2.5, 2.5, 3.75, np.nan, 6.25,
                                                 7.5, 7.5, 8., np.nan])


def test_moving_average_dt_index():
    """
    Test moving average filter with datetimeindex.
    """
    test_jd = pd.date_range(start='2000-01-01', periods=10, freq='D')
    test_data = np.array(
        [1, 2, 3, 4, -999999.0, 6, 7, 8, 9, np.nan], dtype=np.double)

    ser = pd.Series(test_data, index=test_jd)

    filtered = filtering.moving_average(ser, window_size=5)

    np.testing.assert_allclose(filtered.values, [2., 2.5, 2.5, 3.75, np.nan, 6.25,
                                                 7.5, 7.5, 8., np.nan])


def test_moving_average_size_1():
    """
    Test moving average filter with input size 1.
    """
    test_jd = np.arange(1, dtype=np.double)
    test_data = np.array(
        [1], dtype=np.double)

    ser = pd.Series(test_data, index=test_jd)

    filtered = filtering.moving_average(ser, window_size=5)

    np.testing.assert_allclose(filtered.values, [1.])


def test_moving_average_fillna():
    """
    Test moving average filter with datetimeindex.
    """
    test_jd = pd.date_range(start='2000-01-01', periods=12, freq='D')
    test_data = np.array(
        [1, 2, 3, 4, np.nan, np.nan, 8, 9, 10, np.nan, np.nan, np.nan], dtype=np.double)

    ser = pd.Series(test_data, index=test_jd)

    filtered = filtering.moving_average(ser, window_size=5, fillna=True)

    np.testing.assert_allclose(filtered.values, [2., 2.5, 2.5, 3.0, 5.0, 7.0, 9.0, 9.0, 9.0, 9.5, 10, np.nan])


def test_moving_average_min_observations():
    """
    Test moving average filter with datetimeindex.
    """
    test_jd = pd.date_range(start='2000-01-01', periods=12, freq='D')
    test_data = np.array(
        [1, 2, 3, 4, np.nan, np.nan, 8, 9, 10, np.nan, np.nan, np.nan], dtype=np.double)

    ser = pd.Series(test_data, index=test_jd)

    filtered = filtering.moving_average(ser, window_size=5, fillna=True, min_obs=3)

    np.testing.assert_allclose(filtered.values, [2., 2.5, 2.5, 3.0, 5.0, 7.0, 9.0, 9.0, 9.0, np.nan, np.nan, np.nan])

    filtered = filtering.moving_average(ser, window_size=5, fillna=True, min_obs=4)

    np.testing.assert_allclose(filtered.values, [np.nan, 2.5, 2.5, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan])


# Copyright (c) 2014, Vienna University of Technology (TU Wien), Department
# of Geodesy and Geoinformation (GEO).
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the Vienna University of Technology,
#      Department of Geodesy and Geoinformation nor the
#      names of its contributors may be used to endorse or promote products
#      derived from this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL VIENNA UNIVERSITY OF TECHNOLOGY,
# DEPARTMENT OF GEODESY AND GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Author: Christoph Paulik christoph.paulik@geo.tuwien.ac.at
# Creation date: 2014-06-30


"""
Module provides grouping functions that can be used together with pandas
to create a few strange timegroupings like e.g. decadal products were
there are three products per month with timestamps on the 10th 20th and last
of the month
"""

import pandas as pd
import numpy as np
from datetime import date
import calendar


def group_by_day_bin(df, bins=[1, 11, 21, 32], start=False,
                     dtindex=None):
    """
    Calculates timegroups for a given daterange. Groups are from day
    1-10,
    11-20,
    21-last day of each month.

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame with DateTimeIndex for which the grouping should be
        done
    bins : list, optional
        bins in day of the month, default is for dekadal grouping
    start : boolean, optional
        if set to True the start of the bin will be the timestamp for each
        observations
    dtindex : pandas.DatetimeIndex, optional
        precomputed DatetimeIndex that should be used for resulting
        groups, useful for processing of numerous datasets since it does not
        have to be computed for every call

    Returns
    -------
    grouped : pandas.core.groupby.DataFrameGroupBy
        DataFrame groupby object according the the day bins
        on this object functions like sum() or mean() can be
        called to get the desired aggregation.
    dtindex : pandas.DatetimeIndex
        returned so that it can be reused if possible
    """

    dekads = np.digitize(df.index.day, bins)
    if dtindex is None:
        dtindex = grp_to_datetimeindex(dekads, bins, df.index, start=start)
    grp = pd.DataFrame(df.values, columns=df.columns, index=dtindex)
    return grp.groupby(level=0), dtindex


def grp_to_datetimeindex(grps, bins, dtindex, start=False):
    """Makes a datetimeindex that has for each entry the timestamp
    of the bin beginning or end this entry belongs to.

    Parameters
    ----------
    grps : numpy.array
        group numbers made by np.digitize(data, bins)
    bins : list
        bin start values e.g. [0,11,21] would be two bins one with
        values 0<=x<11 and the second one with 11<=x<21
    dtindex : pandas.DatetimeIndex
        same length as grps, gives the basis datetime for each group
    start : boolean, optional
        if set to True the start of the bin will be the timestamp for each
        observations

    Returns
    -------
    grpdt : pd.DatetimeIndex
        Datetimeindex where every date is the end of the bin the datetime
        ind the input dtindex belongs to
    """

    dtlist = []
    offset = 1
    index_offset = 0
    # select previous bin and do not subtract a day if start is set to True
    if start:
        offset = 0
        index_offset = -1

    for i, el in enumerate(dtindex):
        _, max_day_month = calendar.monthrange(el.year, el.month)

        dtlist.append(date(el.year, el.month, min([bins[grps[i] + index_offset]
                                                   - offset, max_day_month])))

    return pd.DatetimeIndex(dtlist)


def grouped_dates_between(start_date, end_date, bins=[1, 11, 21, 32], start=False):
    """
    Between a start and end date give all dates that represent a bin
    See test for example.

    Parameters
    ----------
    start_date: date
        start date
    end_date: date
        end date
    bins: list, optional
        bin start values as days in a month e.g. [0,11,21] would be two bins one with
        values 0<=x<11 and the second one with 11<=x<21
    start: boolean, optional
        if True the start of the bins is the representative date

    Returns
    -------
    tstamps : list of datetimes
        list of representative dates between start and end date
    """

    daily = pd.date_range(start_date, end_date, freq='D')
    fake_data = pd.DataFrame(np.arange(len(daily)), index=daily)
    grp, dtindex = group_by_day_bin(fake_data, bins=bins, start=start)
    tstamps = grp.sum().index.to_pydatetime().tolist()

    return tstamps



# -*- coding: utf-8 -*-

"""
Test functions from the df_metrics module that applies metrics to combinations
of columns from a data frame.
"""

import numpy as np
import pandas as pd
import pytesmo.df_metrics as df_metrics
from pytesmo.metrics import bias, pearsonr
import pytest
from scipy import stats

def test_n_combinations():
    coll = [1,2,3,4]
    combs = df_metrics.n_combinations(coll, n=2, must_include=[1], permutations=False)
    assert combs == [(1,2), (1,3), (1,4)]

    coll = [1, 2, 3, 4]
    combs = df_metrics.n_combinations(coll, n=3, permutations=False)
    assert combs == [(1, 2, 3), (1, 2, 4), (1, 3, 4), (2, 3, 4)]

def test_apply():

    df = pd.DataFrame(index=pd.date_range(start='2000-01-01', end='2000-12-31', freq='D'),
                      data={'ds0': np.repeat(0, 366), 'ds1': np.repeat(1, 366)})
    df.loc[df.index[np.random.choice(range(366), 10)], 'ds0'] = np.nan
    df.loc[df.index[np.random.choice(range(366), 10)], 'ds1'] = np.nan
    with pytest.deprecated_call():
        bias_matrix_old = df_metrics.pairwise_apply(df, bias)
    bias_matrix_new = df_metrics.nwise_apply(df, bias,  n=2, as_df=True)
    r_matrix_new = df_metrics.nwise_apply(df, stats.pearsonr,  n=2, as_df=True)
    assert bias_matrix_old.equals(bias_matrix_new)

    # check if dict implementation and matrix implementation have same result
    bias_new = df_metrics.nwise_apply(df, bias,  n=2, as_df=False)
    for i, v in bias_new.items():
        assert bias_matrix_new.loc[i] == v

def test_dict_to_namedtuple():
    d = {'a': 1, 'b': 2}
    d_named = df_metrics._dict_to_namedtuple(d, 'name')
    assert d_named._fields == ('a', 'b')
    assert type(d_named).__name__ == 'name'

if __name__ == '__main__':
    test_apply()
